/* $Id: common_3dnow.h,v 1.5 1999/05/01 00:46:51 holger Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
 * 3DNow! optimizations contributed by
 * Holger Waechtler <holger@akaflieg.extern.tu-berlin.de>
 */


#ifndef _common_3dnow_h
#define _common_3dnow_h


/*
         returns 1, if 3Dnow! commands are supported by CPU.
*/
extern int hardware_supports_3dnow ();



/*----------------------------------------------------------------------------
    These macros load the operands of a transform function into the reg's:

   static void transform_3dnow_points3_3d_no_rot_raw ( GLvector4f *to_vec,
                        const GLmatrix *mat,
                        const GLvector4f *from_vec,
                        const GLubyte *mask,
                        const GLubyte flag )

   would mean:
         ptr to source vector             from_vec->start  in %%eax
         ptr to matrix                    mat->m           in %%ecx
         ptr to dst vector                to_vec->start    in %%edx
         number of vectors to transform   from_vec->count  in %%esi
         stride                           from_vec->stride in %%edi
  ----------------------------------------------------------------------------*/

#define  BEGIN_XFORM_3DNOW_ASM                                                 \
__asm__ (                                                                     "\
.balign        32                                                            \n\
                                                                             \n\
    pushl      %%edi                                                         \n\
    pushl      %%esi                                                         \n\
                                                                             \n\
    movl       %0, %%edx                                                     \n\
    movl       %1, %%ecx                                                     \n\
    movl       %3, %%esi                                                     \n\
    movl       %4, %%edi                                                     \n\
    movl       %2, %%eax                                                     \n\
                                                                             \n\
    femms                                                                    \n"
 



#define  END_XFORM_3DNOW_ASM                                                  "\
    femms                                                                    \n\
                                                                             \n\
    popl       %%esi                                                         \n\
    popl       %%edi                                                         \n\
                                                                              "\
::"g" (to_vec->start),                                                         \
"g" (mat->m),                                                                  \
"g" (from_vec->start),                                                         \
"g" (((int) from_vec->count)),                                                 \
"g" (((int) from_vec->stride))                                                 \
:"memory"); 


#define  BEGIN_XFORM_3DNOW_ASM_MASKED                                          \
__asm__ (                                                                     "\
.balign        32                                                            \n\
                                                                             \n\
    pushl      %%esi                                                         \n\
    pushl      %%edi                                                         \n\
    pushl      %%ebx                                                         \n\
    pushl      %%ebp                                                         \n\
                                                                             \n\
    movl       %0, %%edx                                                     \n\
    movl       %1, %%ecx                                                     \n\
    movl       %3, %%esi                  # count                            \n\
    movl       %4, %%edi                  # input stride                     \n\
    movl       %2, %%eax                                                     \n\
    movl       36(%%esp), %%ebp           # clipmask                         \n\
    movb       40(%%esp), %%bl            # clip flag                        \n\
                                                                             \n\
    femms                                                                    \n"
 



#define  END_XFORM_3DNOW_ASM_MASKED                                           "\
    femms                                                                    \n\
                                                                             \n\
    popl       %%ebp                                                         \n\
    popl       %%ebx                                                         \n\
    popl       %%edi                                                         \n\
    popl       %%esi                                                         \n\
                                                                              "\
::"g" (to_vec->start),                                                         \
"g" (mat->m),                                                                  \
"g" (from_vec->start),                                                         \
"g" (((int) from_vec->count)),                                                 \
"g" (((int) from_vec->stride))                                                 \
:"memory"); 



#endif

