/* $Id: dotprod_3dnow_masked.h,v 1.2 1999/04/16 01:02:00 brianp Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
 * 3DNow! optimizations contributed by
 * Holger Waechtler <holger@akaflieg.extern.tu-berlin.de>
 */


/*----------------------------------------------------------------------
     r = v0 * p0 + v1 * p1 + v2 * p2 + v3 * p3
  ---------------------------------------------------------------------- */
static void dotprod_3dnow_vec4_masked ( GLvector4f *out_vec,
                 GLuint elt,
                 const GLvector4f *coord_vec,
                 const GLfloat plane[4],
                 const GLubyte mask[])
{
GLfloat*	out = (out_vec->start + elt); 
out_vec->count = coord_vec->count;
__asm__ (
"    femms                                                                  \n"
"                                                                           \n"
".balign        32                                                          \n"
"                                                                           \n"
"    cmp        $0,       %%edx                                             \n"
"    jna        3f                     # nothing to do                      \n"
"                                                                           \n"
"    pushl      %%ebp                                                       \n"
"    movl       24(%%esp),%%ebp        # mask pointer                       \n"
"                                                                           \n"
"    movq       (%%ecx),  %%mm0        # p1               | p0              \n"
"    movq       8(%%ecx), %%mm1        # p3               | p2              \n"
"                                                                           \n"
"1:                                                                         \n"
"    testb      $0,       %%ebp        # clipmask [i] != 0 ??               \n"
"    jnz        2f                                                          \n"
"                                                                           \n"
"    movq       (%%eax),  %%mm2        # x1               | x0              \n"
"    movq       8(%%eax), %%mm3        # x3               | x2              \n"
"                                                                           \n"
"    pfmul      %%mm0,    %%mm2        # x1*p1            | x0*p0           \n"
"    pfmul      %%mm1,    %%mm3        # x3*p3            | x2*p2           \n"
"                                                                           \n"
"    pfadd      %%mm2,    %%mm3        # x1*p1+x3*p3      | x0*p0+x2*p2     \n"
"    pfacc      %%mm3,    %%mm3        # x0*p0+x1*p1+x2*p2+x3*p3            \n"
"                                                                           \n"
"    movd       %%mm3,    (%%ebx)      # write r                            \n"
"2:  addl       %4,       %%eax        # next vector                        \n"
"                                                                           \n"
"    addl       %5,       %%ebx        # next r                             \n"
"    incl       %%ebp                                                       \n"
"                                                                           \n"
"    decl       %%edx                                                       \n"
"    ja         1b                                                          \n"
"3:                                                                         \n"
"    femms                                                                  \n"
"    popl       %%ebp                                                       \n"
"                                                                             "
:"=b" (out) 
:"a" (coord_vec->start), 
"c" (&(plane[0])), 
"d" (coord_vec->count), 
"g" (coord_vec->stride), 
"g" (out_vec->stride)
:"memory");
}




/*----------------------------------------------------------------------
     r = v0 * p0 + v1 * p1 + v2 * p2 + p3
  ---------------------------------------------------------------------- */
static void dotprod_3dnow_vec3_masked ( GLvector4f *out_vec,
                 GLuint elt,
                 const GLvector4f *coord_vec,
                 const GLfloat plane[4],
                 const GLubyte mask[])
{
GLfloat*	out = (out_vec->start + elt); 
out_vec->count = coord_vec->count;
__asm__ (
"    femms                                                                  \n"
"                                                                           \n"
".balign        32                                                          \n"
"                                                                           \n"
"    cmp        $0,       %%edx                                             \n"
"    jna        3f                     # nothing to do                      \n"
"                                                                           \n"
"    pushl      %%ebp                                                       \n"
"    movl       24(%%esp),%%ebp        # mask pointer                       \n"
"                                                                           \n"
"    movq       (%%ecx),  %%mm0        # p1               | p0              \n"
"    movd       8(%%ecx), %%mm1        #                  | p2              \n"
"    movd       12(%%ecx),%%mm2        #                  | p3              \n"
"                                                                           \n"
"1:                                                                         \n"
"    testb      $0,       %%ebp        # clipmask [i] != 0 ??               \n"
"    jnz        2f                                                          \n"
"                                                                           \n"
"    movq       (%%eax),  %%mm3        # x1               | x0              \n"
"    movd       8(%%eax), %%mm4        #                  | x2              \n"
"                                                                           \n"
"    pfmul      %%mm1,    %%mm4        #                  | p2*x2           \n"
"    pfacc      %%mm2,    %%mm4        # p3               | p2*x2           \n"
"                                                                           \n"
"    pfmul      %%mm0,    %%mm3        # p1*x1            | p0*x0           \n"
"    pfadd      %%mm3,    %%mm4        # p3+p1*x1         | p2*x2+x0*x0     \n"
"                                                                           \n"
"    pfacc      %%mm4,    %%mm4        # p0*x0+p1*x1+p2*x2+p3               \n"
"    movd       %%mm4,    (%%ebx)      # write r                            \n"
"                                                                           \n"
"2:  addl       %4,       %%eax        # next vector                        \n"
"    addl       %5,       %%ebx        # next r                             \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%edx                                                       \n"
"                                                                           \n"
"    ja         1b                                                          \n"
"3:                                                                         \n"
"    femms                                                                  \n"
"                                                                             "
:"=b" (out) 
:"a" (coord_vec->start), 
"c" (&(plane[0])), 
"d" (coord_vec->count), 
"g" (coord_vec->stride), 
"g" (out_vec->stride)
:"memory");
}



/*----------------------------------------------------------------------
     r = v0 * p0 + v1 * p1 + p3
  ---------------------------------------------------------------------- */
static void dotprod_3dnow_vec2_masked ( GLvector4f *out_vec,
                 GLuint elt,
                 const GLvector4f *coord_vec,
                 const GLfloat plane[4],
                 const GLubyte mask[])
{
GLfloat*	out = (out_vec->start + elt); 
out_vec->count = coord_vec->count;
__asm__ (
"    femms                                                                  \n"
"                                                                           \n"
".balign        32                                                          \n"
"                                                                           \n"
"    cmp        $0,       %%edx                                             \n"
"    jna        3f                     # nothing to do                      \n"
"                                                                           \n"
"    pushl      %%ebp                                                       \n"
"    movl       24(%%esp),%%ebp        # mask pointer                       \n"
"                                                                           \n"
"    movq       (%%ecx),  %%mm0        # p1               | p0              \n"
"    movd       12(%%ecx),%%mm2        #                  | p3              \n"
"                                                                           \n"
"1:                                                                         \n"
"    testb      $0,       %%ebp        # clipmask [i] != 0 ??               \n"
"    jnz        2f                                                          \n"
"                                                                           \n"
"    movq       (%%eax),  %%mm3        # x1               | x0              \n"
"    pfmul      %%mm0,    %%mm3        # p1*x1            | p0*x0           \n"
"                                                                           \n"
"    pfadd      %%mm2,    %%mm3        # p1*x1            | p0*x0+p3        \n"
"    pfacc      %%mm4,    %%mm4        # p0*x0+p1*x1+p3                     \n"
"                                                                           \n"
"    movd       %%mm4,    (%%ebx)      # write r                            \n"
"2:  addl       %4,       %%eax        # next vector                        \n"
"                                                                           \n"
"    addl       %5,       %%ebx        # next r                             \n"
"    incl       %%ebp                  # next clipmask                      \n"
"                                                                           \n"
"    decl       %%edx                                                       \n"
"    ja         1b                                                          \n"
"3:                                                                         \n"
"    femms                                                                  \n"
"                                                                             "
:"=b" (out) 
:"a" (coord_vec->start), 
"c" (&(plane[0])), 
"d" (coord_vec->count), 
"g" (coord_vec->stride), 
"g" (out_vec->stride)
:"memory");
}



static void gl_init_3dnow_dotprod_masked ()
{
   gl_dotprod_tab [1][2] = dotprod_3dnow_vec2_masked;
   gl_dotprod_tab [1][3] = dotprod_3dnow_vec3_masked;
   gl_dotprod_tab [1][4] = dotprod_3dnow_vec4_masked;
}

