/* $Id: dotprod_3dnow_raw.h,v 1.3 1999/04/16 01:02:00 brianp Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
 * 3DNow! optimizations contributed by
 * Holger Waechtler <holger@akaflieg.extern.tu-berlin.de>
 */

/*
 *  We calculate two dotproducts together.
 */


/*----------------------------------------------------------------------
     r = v0 * p0 + v1 * p1 + v2 * p2 + v3 * p3
  ---------------------------------------------------------------------- */
static void dotprod_3dnow_vec4_raw ( GLvector4f *out_vec,
                 GLuint elt,
                 const GLvector4f *coord_vec,
                 const GLfloat plane[4],
                 const GLubyte mask[])
{
GLfloat*	out = (out_vec->start + elt); 
out_vec->count = coord_vec->count;
__asm__ (
"                                                                           \n"
".balign 32                                                                 \n"
"    femms                                                                  \n"
"    movq       (%%ecx),  %%mm0        # p1               | p0              \n"
"                                                                           \n"
"    decl       %%edx                                                       \n"
"    movq       8(%%ecx), %%mm1        # p3               | p2              \n"
"                                                                           \n"
"    jna        2                                                           \n"
"                                                                           \n"
".balign 32                                                                 \n"
"1:                                                                         \n"
"    movq       (%%eax),  %%mm2        # x1               | x0              \n"
"    movq       8(%%eax), %%mm3        # x3               | x2              \n"
"                                                                           \n"
"    addl       %4,       %%eax        # 2nd vector (y)                     \n"
"    pfmul      %%mm0,    %%mm2        # p1*x1            | p0*x0           \n"
"                                                                           \n"
"    pfmul      %%mm1,    %%mm3        # p3*x3            | p2*x2           \n"
"    movq       (%%eax),  %%mm4        # y1               | y0              \n"
"                                                                           \n"
"    movq       8(%%eax), %%mm5        # y3               | y2              \n"
"    pfmul      %%mm0,    %%mm4        # p1*y1            | p0*y0           \n"
"                                                                           \n"
"    pfmul      %%mm1,    %%mm5        # p3*y3            | p2*y2           \n"
"    pfadd      %%mm2,    %%mm3        # p1*x1+p3*x3      | p0*x0+p2*x2     \n"
"                                                                           \n"
"    pfadd      %%mm4,    %%mm5        # dto. (2nd vector)                  \n"
"    addl       %4,       %%eax        # next v                             \n"
"                                                                           \n"
"    pfacc      %%mm5,    %%mm3        # %%mm3: result    | result (2nd)    \n"
"    movd       %%mm3,    (%%ebx)      # write r                            \n"
"                                                                           \n"
"    addl       %5,       %%ebx        # next r                             \n"
"    punpckhdq  %%mm3,    %%mm3        # cp r -> low 32 Bit of %%mm5        \n"
"                                                                           \n"
"    movd       %%mm3,    (%%ebx)      # write r                            \n"
"    addl       %5,       %%ebx        # next r                             \n"
"                                                                           \n"
"    subl       $2, %%edx                                                   \n"
"    ja         1                                                           \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    jne        3                                                           \n"
"    movq       (%%eax),  %%mm2        # x1               | x0              \n"
"                                                                           \n"
"    movq       8(%%eax), %%mm3        # x3               | x2              \n"
"    pfmul      %%mm0,    %%mm2        # x1*p1            | x0*p0           \n"
"                                                                           \n"
"    pfmul      %%mm1,    %%mm3        # x3*p3            | x2*p2           \n"
"    pfadd      %%mm2,    %%mm3        # x1*p1+x3*p3      | x0*p0+x2*p2     \n"
"                                                                           \n"
"    pfacc      %%mm2,    %%mm3        # x0*p0+x1*p1+x2*p2+x3*p3            \n"
"                                                                           \n"
"    movd       %%mm3,    (%%ebx)      # write r                            \n"
"                                                                           \n"
".balign 32                                                                 \n"
"3:                                                                         \n"
"    femms                                                                  \n"
"                                                                             "
:"=b" (out) 
:"a" (coord_vec->start), 
"c" (&(plane[0])), 
"d" (coord_vec->count), 
"g" (coord_vec->stride), 
"g" (out_vec->stride)
:"memory");
}




/*----------------------------------------------------------------------
     r = v0 * p0 + v1 * p1 + v2 * p2 + p3
  ---------------------------------------------------------------------- */
static void dotprod_3dnow_vec3_raw ( GLvector4f *out_vec,
                 GLuint elt,
                 const GLvector4f *coord_vec,
                 const GLfloat plane[4],
                 const GLubyte mask[])
{
GLfloat*	out = (out_vec->start + elt); 
out_vec->count = coord_vec->count;
__asm__ (
"                                                                           \n"
".balign 32                                                                 \n"
"    femms                                                                  \n"
"    movq       (%%ecx),  %%mm0        # p1               | p0              \n"
"                                                                           \n"
"    movq       8(%%ecx), %%mm1        # p3               | p2              \n"
"    movq       %%mm1,    %%mm5        # p3               | p2              \n"
"                                                                           \n"
"    decl       %%edx                                                       \n"
"    punpckldq  %%mm5,    %%mm5        # p2               | p2              \n"
"                                                                           \n"
"    movq       %%mm1,    %%mm6        # p3               | p2              \n"
"    punpckhdq  %%mm6,    %%mm6        # p3               | p3              \n"
"                                                                           \n"
"    jna        2                                                           \n"
"                                                                           \n"
".balign 32                                                                 \n"
"1:                                                                         \n"
"    movq       (%%eax),  %%mm2        # x1               | x0              \n"
"    movd       8(%%eax), %%mm4        #                  | x2              \n"
"                                                                           \n"
"    addl       %4,       %%eax        # 2nd vector y                       \n"
"    movq       (%%eax),  %%mm3        # y1               | y0              \n"
"                                                                           \n"
"    movd       8(%%eax), %%mm7        #                  | y2              \n"
"                                                                           \n"
"    psllq      $32,      %%mm7        # y2               |                 \n"
"                                                                           \n"
"    por        %%mm7,    %%mm4        # y2               | x2              \n"
"    addl       %4,       %%eax        # next vector                        \n"
"                                                                           \n"
"    pfmul      %%mm0,    %%mm2        # p1*x1            | p0*x0           \n"
"                                                                           \n"
"    pfmul      %%mm0,    %%mm3        # p1*y1            | p0*y0           \n"
"                                                                           \n"
"    pfmul      %%mm5,    %%mm4        # p2*x2            | p2*y2           \n"
"    pfacc      %%mm3,    %%mm2        # p0*x0+p1*x1      | p0*y0+p1*y1     \n"
"                                                                           \n"
"    pfadd      %%mm4,    %%mm2        # p0*x0+p1*x1+p2*x2| sum (p0*y0)     \n"
"    pfadd      %%mm6,    %%mm2        # p0*x0+p1*x1+p2*x2+p3 | dto y[]     \n"
"                                                                           \n"
"    movd       %%mm2,    (%%ebx)      # write r                            \n"
"                                                                           \n"
"    addl       %5,       %%ebx        # next r                             \n"
"    punpckhdq  %%mm2,    %%mm2        # cp r -> low 32 Bit of %%mm5        \n"
"                                                                           \n"
"    movd       %%mm2,    (%%ebx)      # write r                            \n"
"    addl       %5,       %%ebx        # next r                             \n"
"                                                                           \n"
"    subl       $2, %%edx                                                   \n"
"    ja         1                                                           \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    jne        3                                                           \n"
"    movq       (%%eax),  %%mm2        # x1               | x0              \n"
"                                                                           \n"
"    pfmul      %%mm0,    %%mm2        # p1*x1            | p0*x0           \n"
"    movd       8(%%eax), %%mm4        # x3               | x2              \n"
"                                                                           \n"
"    pfmul      %%mm1,    %%mm4        # p2*x3            | p2*x2           \n"
"    pfacc      %%mm3,    %%mm2        # **trash**        | p0*x0+p1*x1     \n"
"                                                                           \n"
"    pfadd      %%mm2,    %%mm4        # p0*x0+p1*x1+p2*x2                  \n"
"    pfadd      %%mm6,    %%mm4        # p0*x0+p1*x1+p2*x2+p3               \n"
"                                                                           \n"
"    movd       %%mm4,    (%%ebx)      # write r                            \n"
"                                                                           \n"
".balign 32                                                                 \n"
"3:                                                                         \n"
"    femms                                                                  \n"
"                                                                             "
:"=b" (out) 
:"a" (coord_vec->start), 
"c" (&(plane[0])), 
"d" (coord_vec->count), 
"g" (coord_vec->stride), 
"g" (out_vec->stride)
:"memory");
}



/*----------------------------------------------------------------------
     r = v0 * p0 + v1 * p1 + p3
  ---------------------------------------------------------------------- */
static void dotprod_3dnow_vec2_raw ( GLvector4f *out_vec,
                 GLuint elt,
                 const GLvector4f *coord_vec,
                 const GLfloat plane[4],
                 const GLubyte mask[])
{
GLfloat*	out = (out_vec->start + elt); 
out_vec->count = coord_vec->count;
__asm__ (
"                                                                           \n"
".balign 32                                                                 \n"
"    femms                                                                  \n"
"    movq       (%%ecx),  %%mm0        # p1              | p0               \n"
"                                                                           \n"
"    movq       8(%%ecx), %%mm1        # p3              | p2               \n"
"    punpckhdq  %%mm1,    %%mm1        # p3              | p3               \n"
"                                                                           \n"
"    decl       %%edx                                                       \n"
"    jna        2                                                           \n"
"                                                                           \n"
".balign 32                                                                 \n"
"1:                                                                         \n"
"    movq       (%%eax),  %%mm2        # x1              | x0               \n"
"    addl       %4,       %%eax        # next vector                        \n"
"                                                                           \n"
"    movq       (%%eax),  %%mm3        # y1              | y0               \n"
"    pfmul      %%mm0,    %%mm2        # x1*p1           | x0*p0            \n"
"                                                                           \n"
"    pfmul      %%mm0,    %%mm3        # y1*p1           | y0*p0            \n"
"    addl       %4,       %%eax        # next vector                        \n"
"                                                                           \n"
"    pfacc      %%mm3,    %%mm2        # x0*p0+x1*p1     | y0*p0+y1*p1      \n"
"    pfadd      %%mm1,    %%mm2        # x0*p0+x1*p1+p3  | y0*p0+y1*p1+p3   \n"
"                                                                           \n"
"    movd       %%mm2,    (%%ebx)      # write r                            \n"
"                                                                           \n"
"    addl       %5,       %%ebx        # next r                             \n"
"    punpckhdq  %%mm2,    %%mm2        # cp r -> low 32 Bit of %%mm5        \n"
"                                                                           \n"
"    movd       %%mm2,    (%%ebx)      # write r                            \n"
"                                                                           \n"
"    addl       %5,       %%ebx        # next r                             \n"
"    subl       $2, %%edx                                                   \n"
"                                                                           \n"
"    ja         1                                                           \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    jne        dprod2_end                                                  \n"
"    movq       (%%eax),  %%mm2        # x1              | x0               \n"
"                                                                           \n"
"    pfmul      %%mm0,    %%mm2        # x1*p1           | x0*p0            \n"
"    pfacc      %%mm0,    %%mm2        #                 | x0*p0+x1*p1      \n"
"                                                                           \n"
"    pfadd      %%mm1,    %%mm2        #                 | x0*p0+x1*p1+p3   \n"
"    movd       %%mm2,    (%%ebx)      # write r                            \n"
"                                                                           \n"
".balign 32                                                                 \n"
"3:                                                                         \n"
"    femms                                                                  \n"
"                                                                             "
:"=b" (out) 
:"a" (coord_vec->start), 
"c" (&(plane[0])), 
"d" (coord_vec->count), 
"g" (coord_vec->stride), 
"g" (out_vec->stride)
:"memory");
}



static void gl_init_3dnow_dotprod_raw ()
{
   gl_dotprod_tab [0][2] = dotprod_3dnow_vec2_raw;
   gl_dotprod_tab [0][3] = dotprod_3dnow_vec3_raw;
   gl_dotprod_tab [0][4] = dotprod_3dnow_vec4_raw;
}

