
#include "common_3dnow.h"
#include "types.h"


void gl_3dnow_transform_normalize_normals_raw ( const GLmatrix *mat,
					      GLfloat scale,
					      const GLvector3f *in,
					      const GLfloat *lengths,
					      const GLubyte mask[],
					      GLvector3f *dest )
{
   (void) scale;
   (void) mask;

   dest->count = in->count;

__asm__ ( 
".balign       32                                                           \n"
"                                                                           \n"
"    cmp        $0,         %3         # count > 0 ??                       \n"
"    je         9f                                                          \n"
"                                                                           \n"
"    pushl      %%ebp                                                       \n"
"    movl       %3,         %%ebp                                           \n"
"    femms                                                                  \n"
"                                                                           \n"
"    pushl      %%ebp                                                       \n"
"    pushl      %0                                                          \n"
"    pushl      %2                     # save counter & pointer for         \n"
"    pushl      %4                     # the normalize pass                 \n"
"                                                                           \n"
"                                                                           \n"
"    movq       (%1),       %%mm3      # m1            | m0                 \n"
"    movq       16(%1),     %%mm4      # m5            | m4                 \n"
"                                                                           \n"
"    movd       8(%1),      %%mm5      #               | m2                 \n"
"    movd       24(%1),     %%mm7      #               | m6                 \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m6            |                    \n"
"    por        %%mm7,      %%mm5      # m6            | m2                 \n"
"                                                                           \n"
"    movq       32(%1),     %%mm6      # m9            | m8                 \n"
"    movd       40(%1),     %%mm7      #               | m10                \n"
"                                                                           \n"
"    movq       (%2),       %%mm0      # x1            | x0                 \n"
"    movd       8(%2),      %%mm2      #               | x2                 \n"
"                                                                           \n"
"1:                                                                         \n"
"    movq       %%mm0,      %%mm1      # x1            | x0                 \n"
"    punpckldq  %%mm2,      %%mm2      # x2            | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm3,      %%mm0      # x1*m1         | x0*m0              \n"
"    addl       $12,        %0         # next r                             \n"
"                                                                           \n"
"    pfmul      %%mm4,      %%mm1      # x1*m5         | x0*m4              \n"
"    pfacc      %%mm1,      %%mm0      # x0*m4+x1*m5   | x0*m0+x1*m1        \n"
"                                                                           \n"
"    pfmul      %%mm5,      %%mm2      # x2*m6         | x2*m2              \n"
"    pfadd      %%mm2,      %%mm0      # x0*m4...+x2*m6| x0*m0+x1*m1+x2*m2  \n"
"                                                                           \n"
"    movq       (%2),       %%mm1      # x1            | x0                 \n"
"    movq       %%mm0,      -12(%0)    # write r0, r1                       \n"
"                                                                           \n"
"    pfmul      %%mm6,      %%mm1      # x1*m9         | x0*m8              \n"
"    movd       8(%2),      %%mm2      #               | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm7,      %%mm2      #               | x2*m10             \n"
"    pfacc      %%mm1,      %%mm1      # *not used*    | x0*m8+x1*m9        \n"
"                                                                           \n"
"    pfadd      %%mm2,      %%mm1      # *not used*    | x0*m8+x1*m9+x2*m10 \n"
"    addl       %4,         %2         # next normal                        \n"
"                                                                           \n"
"    movd       %%mm1,      -4(%0)     # write r2                           \n"
"    movq       (%2),       %%mm0      # x1            | x0                 \n"
"                                                                           \n"
"    movd       8(%2),      %%mm2      #               | x2                 \n"
"    decl       %%ebp                  # decrement normal counter           \n"
"    ja         1b                                                          \n"
"                                                                           \n"
"                                      #  end of transform ---              \n"
"    popl       %4                     #   now normalizing ...              \n"
"    popl       %2                                                          \n"
"    popl       %0                                                          \n"
"    popl       %%ebp                                                       \n"
"                                                                           \n"
"    movq       (%0),       %%mm0      # x1              | x0               \n"
"    movd       8(%0),      %%mm1      #                 | x2               \n"
"                                                                           \n"
"    cmp        $0,         %5         # lengths == 0 ?                     \n"
"    je         5f                     # calculate lengths                  \n"
"                                                                           \n"
"3:                                    # use precalculated lengths          \n"
"    movd       (%5),       %%mm3      #                 | length (x)       \n"
"    pfmul      %%mm3,      %%mm1      #                 | x2 (normalized)  \n"
"                                                                           \n"
"    punpckldq  %%mm3,      %%mm3      # length (x)      | length (x)       \n"
"    pfmul      %%mm3,      %%mm0      # x1 (normalized) | x0 (normalized)  \n"
"                                                                           \n"
"    movq       %%mm0,      (%0)       # write new x0, x1                   \n"
"    movd       %%mm1,      8(%0)      # write new x2                       \n"
"                                                                           \n"
"    addl       %4,         %2         # next normal                        \n"
"    addl       $12,        %0         # next r                             \n"
"                                                                           \n"
"    addl       $4,         %5         # next length                        \n"
"    decl       %%ebp                  # decrement normal counter           \n"
"                                                                           \n"
"    movq       (%0),       %%mm0      # x1              | x0               \n"
"    movd       8(%0),      %%mm1      #                 | x2               \n"
"    ja         3b                                                          \n"
"    jmp        8f                     #  that's all !!                     \n"
"                                                                           \n"
"5:                                    # need to calculate lengths          \n"
"    movq       %%mm0,      %%mm3      # x1              | x0               \n"
"    movq       %%mm1,      %%mm4      #                 | x2               \n"
"                                                                           \n"
"    pfmul      %%mm0,      %%mm3      # x1*x1           | x0*x0            \n"
"    addl       $12,        %0         # next r                             \n"
"                                                                           \n"
"    pfmul      %%mm1,      %%mm4      #                 | x2*x2            \n"
"    pfadd      %%mm4,      %%mm3      #                 | x0*x0+x2*x2      \n"
"                                                                           \n"
"    pfacc      %%mm3,      %%mm3      # x0*x0+...+x2*x2 | x0*x0+x1*x1+x2*x2\n"
"    pfrsqrt    %%mm3,      %%mm5      # 1/sqrt (x0*x0+x1*x1+x2*x2)         \n"
"                                                                           \n"
"    decl       %%ebp                  # decrement normal counter           \n"
"    pfmul      %%mm5,      %%mm0      # x1 (normalized) | x0 (normalized)  \n"
"                                                                           \n"
"    movq       %%mm0,      -12(%0)    # write new x0, x1                   \n"
"    pfmul      %%mm5,      %%mm1      #                 | x2 (normalized)  \n"
"                                                                           \n"
"    movd       %%mm1,      -4(%0)     # write new x2                       \n"
"    movq       (%0),       %%mm0      # x1              | x0               \n"
"                                                                           \n"
"    movd       8(%0),      %%mm1      #                 | x2               \n"
"    ja         5b                                                          \n"
"                                                                           \n"
"8:  femms                                                                  \n"
"    popl       %%ebp                                                       \n"
"9:                                                                         \n"
"                                                                           \n"
::"a" (dest->start),
"c" (mat->inv),
"d" (in->start),
"g" (((int) in->count)),
"g" (((int) in->stride)),
"D" (lengths)
:"memory");
}



void gl_3dnow_transform_normalize_normals_no_rot_raw ( const GLmatrix *mat,
						     GLfloat scale,
						     const GLvector3f *in,
						     const GLfloat *lengths,
						     const GLubyte mask[],
						     GLvector3f *dest )
{
   (void) scale;
   (void) mask;

   dest->count = in->count;

__asm__ (
".balign       32                                                           \n"
"                                                                           \n"
"    cmp        $0,         %3         # count > 0 ??                       \n"
"    je         9f                                                          \n"
"                                                                           \n"
"    pushl      %%ebp                                                       \n"
"    movl       %3,         %%ebp                                           \n"
"    femms                                                                  \n"
"                                                                           \n"
"    movd       (%1),       %%mm0      #               | m0                 \n"
"    movd       20(%1),     %%mm7      #               | m5                 \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m5            |                    \n"
"    por        %%mm7,      %%mm0      # m5            | m0                 \n"
"                                                                           \n"
"    movd       40(%1),     %%mm2      #               | m10                \n"
"    punpckldq  %%mm2,      %%mm2      # m10           | m10                \n"
"                                                                           \n"
"    movq       (%2),       %%mm6      # x1            | x0                 \n"
"    movd       8(%2),      %%mm7      #               | x2                 \n"
"                                                                           \n"
"    cmp        $0,         %5         # lengths == 0 ?                     \n"
"    je         5f                     # need to calculate lengths          \n"
"                                                                           \n"
"    movd       (%5),       %%mm3      #                 | length (x)       \n"
"                                                                           \n"
"1:                                    # use precalculated lengths          \n"
"    pfmul      %%mm0,      %%mm6      # x1*m5         | x0*m0              \n"
"    addl       %4,         %2         # next normal                        \n"
"                                                                           \n"
"    pfmul      %%mm2,      %%mm7      #               | x2*m10             \n"
"    addl       $12,        %0         # next r                             \n"
"                                                                           \n"
"    pfmul      %%mm3,      %%mm7      #                 | x2 (normalized)  \n"
"    punpckldq  %%mm3,      %%mm3      # length (x)      | length (x)       \n"
"                                                                           \n"
"    addl       $4,         %5         # next length                        \n"
"    pfmul      %%mm3,      %%mm6      # x1 (normalized) | x0 (normalized)  \n"
"                                                                           \n"
"    decl       %%ebp                  # decrement normal counter           \n"
"    movq       %%mm6,      -12(%0)    # write r0, r1                       \n"
"                                                                           \n"
"    movd       %%mm7,      -4(%0)     # write r2                           \n"
"    movd       (%5),       %%mm3      #                 | length (x)       \n"
"                                                                           \n"
"    movq       (%2),       %%mm6      # x1              | x0               \n"
"    movd       8(%2),      %%mm7      #                 | x2               \n"
"    ja         1b                                                          \n"
"                                                                           \n"
"    jmp        8f                     #  that's all !!                     \n"
"                                                                           \n"
"5:                                    # need to calculate lengths          \n"
"    pfmul      %%mm0,      %%mm6      # x1*m5           | x0*m0            \n"
"    addl       $12,        %0         # next r                             \n"
"                                                                           \n"
"    pfmul      %%mm2,      %%mm7      #                 | x2*m10           \n"
"    movq       %%mm6,      %%mm3      # x1 (transformed)| x0 (transformed) \n"
"                                                                           \n"
"    movq       %%mm7,      %%mm4      #                 | x2 (transformed) \n"
"    pfmul      %%mm6,      %%mm3      # x1*x1           | x0*x0            \n"
"                                                                           \n"
"    pfmul      %%mm7,      %%mm4      #                 | x2*x2            \n"
"    pfacc      %%mm3,      %%mm3      # **not used**    | x0*x0+x1*x1      \n"
"                                                                           \n"
"    pfadd      %%mm4,      %%mm3      #                 | x0*x0+x1*x1+x2*x2\n"
"    addl       %4,         %2         # next normal                        \n"
"                                                                           \n"
"    pfrsqrt    %%mm3,      %%mm5      # 1/sqrt (x0*x0+x1*x1+x2*x2)         \n"
"    decl       %%ebp                  # decrement normal counter           \n"
"                                                                           \n"
"    pfmul      %%mm5,      %%mm6      # x1 (normalized) | x0 (normalized)  \n"
"    movq       %%mm6,      -12(%0)    # write r0, r1                       \n"
"                                                                           \n"
"    pfmul      %%mm5,      %%mm7      #                 | x2 (normalized)  \n"
"    movd       %%mm7,      -4(%0)     # write r2                           \n"
"                                                                           \n"
"    movq       (%2),       %%mm6      # x1              | x0               \n"
"    movd       8(%2),      %%mm7      #                 | x2               \n"
"                                                                           \n"
"    ja         5b                                                          \n"
"                                                                           \n"
"8:  femms                                                                  \n"
"    popl       %%ebp                                                       \n"
"9:                                                                         \n"
"                                                                           \n"
::"a" (dest->start),
"c" (mat->inv),
"d" (in->start),
"g" (((int) in->count)),
"g" (((int) in->stride)),
"D" (lengths)
:"memory");
}




/*----------------------------------------------------------------
	 out[i][0] = ux * m0 * scale;
	 out[i][1] =           uy * m5 * scale;
	 out[i][2] =                     uz * m10 * scale;
  --------------------------------------------------------------*/
void gl_3dnow_transform_rescale_normals_no_rot_raw (const GLmatrix *mat,
					  GLfloat scale,
					  const GLvector3f *in,
					  const GLfloat *lengths,
					  const GLubyte mask[],
					  GLvector3f *dest )
{
   (void) lengths;
   (void) mask;
   dest->count = in->count;

__asm__ (
".balign       32                                                           \n"
"                                                                           \n"
"    cmp        $0,         %3                                              \n"
"    je         4f                                                          \n"
"                                                                           \n"
"    femms                                                                  \n"
"                                                                           \n"
"    movd       %5,         %%mm6      #               | scale              \n"
"    punpckldq  %%mm6,      %%mm6      # scale         | scale              \n"
"                                                                           \n"
"    movd       (%1),       %%mm0      #               | m0                 \n"
"    movd       20(%1),     %%mm7      #               | m5                 \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m5            |                    \n"
"    por        %%mm7,      %%mm0      # m5            | m0                 \n"
"                                                                           \n"
"    pfmul      %%mm6,      %%mm0      # scale*m5      | scale*m0           \n"
"    movd       40(%1),     %%mm2      #               | m10                \n"
"                                                                           \n"
"    movq       (%2),       %%mm4      # x1            | x0                 \n"
"    pfmul      %%mm6,      %%mm2      #               | scale*m10          \n"
"                                                                           \n"
"    movd       8(%2),      %%mm5      #               | x2                 \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:  pfmul      %%mm0,      %%mm4      # x1*m5         | x0*m0              \n"
"    addl       %4,         %2         # next normal                        \n"
"                                                                           \n"
"    pfmul      %%mm2,      %%mm5      #               | x2*m10             \n"
"    addl       $12,        %0         # next r                             \n"
"                                                                           \n"
"    decl       %3                     # decrement normal counter           \n"
"    movq       %%mm4,      -12(%0)    # write r0, r1                       \n"
"                                                                           \n"
"    movd       %%mm5,      -4(%0)     # write r2                           \n"
"    movq       (%2),       %%mm4      # x1            | x0                 \n"
"                                                                           \n"
"    movd       8(%2),      %%mm5      #               | x2                 \n"
"    ja         2b                     # cnt > 0 ? -> process next normal   \n"
"                                                                           \n"
"    femms                                                                  \n"
"4:                                                                         \n"
"                                                                           \n"
::"a" (dest->start),
"c" (mat->inv),
"d" (in->start),
"D" (((int) in->count)),
"g" (((int) in->stride)),
"g" (scale)
:"memory");
}



/*----------------------------------------------------------------
	 out[i][0] = scale * (ux * m0 + uy * m1 + uz * m2);
	 out[i][1] = scale * (ux * m4 + uy * m5 + uz * m6);
	 out[i][2] = scale * (ux * m8 + uy * m9 + uz * m10);
  --------------------------------------------------------------*/
void gl_3dnow_transform_rescale_normals_raw ( const GLmatrix *mat,
					    GLfloat scale,
					    const GLvector3f *in,
					    const GLfloat *lengths,
					    const GLubyte mask[],
					    GLvector3f *dest )
{
   (void) lengths;
   (void) mask;

   dest->count = in->count;

__asm__ (
".balign       32                                                           \n"
"                                                                           \n"
"    cmp        $0,         %3         # count > 0 ??                       \n"
"    je         9f                                                          \n"
"                                                                           \n"
"    femms                                                                  \n"
"                                                                           \n"
"    movq       (%1),       %%mm3      # m1            | m0                 \n"
"                                                                           \n"
"    movq       16(%1),     %%mm4      # m5            | m4                 \n"
"    movd       8(%1),      %%mm5      #               | m2                 \n"
"                                                                           \n"
"    movd       24(%1),     %%mm7      #               | m6                 \n"
"    movd       %5,         %%mm0      # scale                              \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m6            |                    \n"
"    punpckldq  %%mm0,      %%mm0      # scale         | scale              \n"
"                                                                           \n"
"    por        %%mm7,      %%mm5      # m6            | m2                 \n"
"    pfmul      %%mm0,      %%mm3      # scale*m1      | scale*m0           \n"
"                                                                           \n"
"    movq       32(%1),     %%mm6      # m9            | m8                 \n"
"    pfmul      %%mm0,      %%mm4      # scale*m5      | scale*m4           \n"
"                                                                           \n"
"    movd       40(%1),     %%mm7      #               | m10                \n"
"    pfmul      %%mm0,      %%mm5      # scale*m6      | scale*m2           \n"
"                                                                           \n"
"    pfmul      %%mm0,      %%mm6      # scale*m9      | scale*m8           \n"
"    movd       8(%2),      %%mm2      #               | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm0,      %%mm7      #               | scale*m10          \n"
"    movq       (%2),       %%mm0      # x1            | x0                 \n"
"                                                                           \n"
"1:                                                                         \n"
"    movq       %%mm0,      %%mm1      # x1            | x0                 \n"
"    punpckldq  %%mm2,      %%mm2      # x2            | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm3,      %%mm0      # x1*m1         | x0*m0              \n"
"    addl       $12,        %0         # next r                             \n"
"                                                                           \n"
"    pfmul      %%mm4,      %%mm1      # x1*m5         | x0*m4              \n"
"    pfacc      %%mm1,      %%mm0      # x0*m4+x1*m5   | x0*m0+x1*m1        \n"
"                                                                           \n"
"    pfmul      %%mm5,      %%mm2      # x2*m6         | x2*m2              \n"
"    pfadd      %%mm2,      %%mm0      # x0*m4...+x2*m6| x0*m0+x1*m1+x2*m2  \n"
"                                                                           \n"
"    movq       (%2),       %%mm1      # x1            | x0                 \n"
"    movq       %%mm0,      -12(%0)    # write r0, r1                       \n"
"                                                                           \n"
"    pfmul      %%mm6,      %%mm1      # x1*m9         | x0*m8              \n"
"    movd       8(%2),      %%mm2      #               | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm7,      %%mm2      #               | x2*m10             \n"
"    pfacc      %%mm1,      %%mm1      # *not used*    | x0*m8+x1*m9        \n"
"                                                                           \n"
"    pfadd      %%mm2,      %%mm1      # *not used*    | x0*m8+x1*m9+x2*m10 \n"
"    addl       %4,         %2         # next normal                        \n"
"                                                                           \n"
"    movd       %%mm1,      -4(%0)     # write r2                           \n"
"    movq       (%2),       %%mm0      # x1            | x0                 \n"
"                                                                           \n"
"    movd       8(%2),      %%mm2      #               | x2                 \n"
"    decl       %3                     # decrement normal counter           \n"
"    ja         1b                                                          \n"
"                                                                           \n"
"    femms                                                                  \n"
"9:                                                                         \n"
"                                                                           \n"
::"a" (dest->start),
"c" (mat->inv),
"d" (in->start),
"D" (((int) in->count)),
"g" (((int) in->stride)),
"g" (scale)
:"memory");
}




/*----------------------------------------------------------------
	 out[i][0] = ux * m0;
	 out[i][1] =           uy * m5;
	 out[i][2] =                     uz * m10;
  --------------------------------------------------------------*/
void gl_3dnow_transform_normals_no_rot_raw (const GLmatrix *mat,
					  GLfloat scale,
					  const GLvector3f *in,
					  const GLfloat *lengths,
					  const GLubyte mask[],
					  GLvector3f *dest )
{
   (void) scale;
   (void) lengths;
   (void) mask;
   dest->count = in->count;

__asm__ (
".balign       32                                                           \n"
"                                                                           \n"
"    cmp        $0,         %3                                              \n"
"    je         4f                                                          \n"
"                                                                           \n"
"    femms                                                                  \n"
"                                                                           \n"
"    movd       (%1),       %%mm0      #               | m0                 \n"
"    movd       20(%1),     %%mm7      #               | m5                 \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m5            |                    \n"
"    por        %%mm7,      %%mm0      # m5            | m0                 \n"
"                                                                           \n"
"    movd       40(%1),     %%mm2      #               | m10                \n"
"    punpckldq  %%mm2,      %%mm2      # m10           | m10                \n"
"                                                                           \n"
"    movq       (%2),       %%mm4      # x1            | x0                 \n"
"    movd       8(%2),      %%mm5      #               | x2                 \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:  pfmul      %%mm0,      %%mm4      # x1*m5         | x0*m0              \n"
"    addl       %4,         %2         # next normal                        \n"
"                                                                           \n"
"    pfmul      %%mm2,      %%mm5      #               | x2*m10             \n"
"    addl       $12,        %0         # next r                             \n"
"                                                                           \n"
"    decl       %3                     # decrement normal counter           \n"
"    movq       %%mm4,      -12(%0)    # write r0, r1                       \n"
"                                                                           \n"
"    movd       %%mm5,      -4(%0)     # write r2                           \n"
"    movq       (%2),       %%mm4      # x1            | x0                 \n"
"                                                                           \n"
"    movd       8(%2),      %%mm5      #               | x2                 \n"
"    ja         2b                     # cnt > 0 ? -> process next normal   \n"
"                                                                           \n"
"    femms                                                                  \n"
"4:                                                                         \n"
"                                                                             "
::"a" (dest->start),
"c" (mat->inv),
"d" (in->start),
"D" (((int) in->count)),
"g" (((int) in->stride))
:"memory");
}


/*----------------------------------------------------------------
	 out[i][0] = ux * m0 + uy * m1 + uz * m2;
	 out[i][1] = ux * m4 + uy * m5 + uz * m6;
	 out[i][2] = ux * m8 + uy * m9 + uz * m10;
  --------------------------------------------------------------*/
void gl_3dnow_transform_normals_raw ( const GLmatrix *mat,
				    GLfloat scale,
				    const GLvector3f *in,
				    const GLfloat *lengths,
				    const GLubyte mask[],
				    GLvector3f *dest )
{
   (void) scale;
   (void) lengths;
   (void) mask;

   dest->count = in->count;

__asm__ (
".balign       32                                                           \n"
"                                                                           \n"
"    cmp        $0,         %3         # count > 0 ??                       \n"
"    je         9f                                                          \n"
"                                                                           \n"
"    femms                                                                  \n"
"                                                                           \n"
"    movq       (%1),       %%mm3      # m1            | m0                 \n"
"    movq       16(%1),     %%mm4      # m5            | m4                 \n"
"                                                                           \n"
"    movd       8(%1),      %%mm5      #               | m2                 \n"
"    movd       24(%1),     %%mm7      #               | m6                 \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m6            |                    \n"
"    por        %%mm7,      %%mm5      # m6            | m2                 \n"
"                                                                           \n"
"    movq       32(%1),     %%mm6      # m9            | m8                 \n"
"    movd       40(%1),     %%mm7      #               | m10                \n"
"                                                                           \n"
"    movq       (%2),       %%mm0      # x1            | x0                 \n"
"    movd       8(%2),      %%mm2      #               | x2                 \n"
"                                                                           \n"
"1:                                                                         \n"
"    movq       %%mm0,      %%mm1      # x1            | x0                 \n"
"    punpckldq  %%mm2,      %%mm2      # x2            | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm3,      %%mm0      # x1*m1         | x0*m0              \n"
"    addl       $12,        %0         # next r                             \n"
"                                                                           \n"
"    pfmul      %%mm4,      %%mm1      # x1*m5         | x0*m4              \n"
"    pfacc      %%mm1,      %%mm0      # x0*m4+x1*m5   | x0*m0+x1*m1        \n"
"                                                                           \n"
"    pfmul      %%mm5,      %%mm2      # x2*m6         | x2*m2              \n"
"    pfadd      %%mm2,      %%mm0      # x0*m4...+x2*m6| x0*m0+x1*m1+x2*m2  \n"
"                                                                           \n"
"    movq       (%2),       %%mm1      # x1            | x0                 \n"
"    movq       %%mm0,      -12(%0)    # write r0, r1                       \n"
"                                                                           \n"
"    pfmul      %%mm6,      %%mm1      # x1*m9         | x0*m8              \n"
"    movd       8(%2),      %%mm2      #               | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm7,      %%mm2      #               | x2*m10             \n"
"    pfacc      %%mm1,      %%mm1      # *not used*    | x0*m8+x1*m9        \n"
"                                                                           \n"
"    pfadd      %%mm2,      %%mm1      # *not used*    | x0*m8+x1*m9+x2*m10 \n"
"    addl       %4,         %2         # next normal                        \n"
"                                                                           \n"
"    movd       %%mm1,      -4(%0)     # write r2                           \n"
"    movq       (%2),       %%mm0      # x1            | x0                 \n"
"                                                                           \n"
"    movd       8(%2),      %%mm2      #               | x2                 \n"
"    decl       %3                     # decrement normal counter           \n"
"    ja         1b                                                          \n"
"                                                                           \n"
"    femms                                                                  \n"
"9:                                                                         \n"
"                                                                             "
::"a" (dest->start),
"c" (mat->inv),
"d" (in->start),
"D" (((int) in->count)),
"g" (((int) in->stride))
:"memory");
}



void gl_3dnow_normalize_normals_raw ( const GLmatrix *mat,
				    GLfloat scale,
				    const GLvector3f *in,
				    const GLfloat *lengths,
				    const GLubyte mask[],
				    GLvector3f *dest )
{
   (void) mat;
   (void) mask;
   (void) scale;
   dest->count = in->count;

__asm__ (
".balign       32                                                           \n"
"                                                                           \n"
"    cmp        $0,         %3         # count > 0 ??                       \n"
"    je         9f                                                          \n"
"                                                                           \n"
"    pushl      %%esi                                                       \n"
"    movl       %3,         %%esi                                           \n"
"    femms                                                                  \n"
"                                                                           \n"
"    movq       (%2),       %%mm0      # x1              | x0               \n"
"    movd       8(%2),      %%mm1      #                 | x2               \n"
"                                                                           \n"
"    cmp        $0,         %5         # lengths == 0 ?                     \n"
"    je         5f                     # calculate lengths                  \n"
"                                                                           \n"
"1:                                    # use precalculated lengths          \n"
"    movd       (%5),       %%mm3      #                 | length (x)       \n"
"    pfmul      %%mm3,      %%mm1      #                 | x2 (normalized)  \n"
"                                                                           \n"
"    punpckldq  %%mm3,      %%mm3      # length (x)      | length (x)       \n"
"    pfmul      %%mm3,      %%mm0      # x1 (normalized) | x0 (normalized)  \n"
"                                                                           \n"
"    movq       %%mm0,      (%0)       # write new x0, x1                   \n"
"    movd       %%mm1,      8(%0)      # write new x2                       \n"
"                                                                           \n"
"    addl       %4,         %2         # next normal                        \n"
"    addl       $12,        %0         # next r                             \n"
"                                                                           \n"
"    addl       $4,         %5         # next length                        \n"
"    decl       %%esi                  # decrement normal counter           \n"
"                                                                           \n"
"    movq       (%2),       %%mm0      # x1              | x0               \n"
"    movd       8(%2),      %%mm1      #                 | x2               \n"
"    ja         1b                                                          \n"
"                                                                           \n"
"    jmp        8f                     #  that's all !!                     \n"
"                                                                           \n"
"5:                                    # need to calculate lengths          \n"
"    movq       %%mm0,      %%mm3      # x1              | x0               \n"
"    addl       %4,         %2         # next normal                        \n"
"                                                                           \n"
"    pfmul      %%mm0,      %%mm3      # x1*x1           | x0*x0            \n"
"    movq       %%mm1,      %%mm4      #                 | x2               \n"
"                                                                           \n"
"    addl       $12,        %0         # next r                             \n"
"    pfmul      %%mm1,      %%mm4      #                 | x2*x2            \n"
"                                                                           \n"
"    pfadd      %%mm4,      %%mm3      #                 | x0*x0+x2*x2      \n"
"    pfacc      %%mm3,      %%mm3      # x0*x0+...+x2*x2 | x0*x0+x1*x1+x2*x2\n"
"                                                                           \n"
"    pfrsqrt    %%mm3,      %%mm5      # 1/sqrt (x0*x0+x1*x1+x2*x2)         \n"
"                                      #   (15 bit precision)               \n"
"    decl       %%esi                  # decrement normal counter           \n"
"                                                                           \n"
"    pfmul      %%mm5,      %%mm0      # x1 (normalized) | x0 (normalized)  \n"
"    movq       %%mm0,      -12(%0)    # write new x0, x1                   \n"
"                                                                           \n"
"    pfmul      %%mm5,      %%mm1      #                 | x2 (normalized)  \n"
"    movd       %%mm1,      -4(%0)     # write new x2                       \n"
"                                                                           \n"
"    movq       (%2),       %%mm0      # x1              | x0               \n"
"    movd       8(%2),      %%mm1      #                 | x2               \n"
"    ja         5b                                                          \n"
"                                                                           \n"
"8:  femms                                                                  \n"
"    popl       %%esi                                                       \n"
"9:                                                                         \n"
"                                                                           \n"
::"a" (dest->start),
"g" (mat->inv),
"c" (in->start),
"g" (((int) in->count)),
"g" (((int) in->stride)),
"d" (lengths)
:"memory");
}


/*----------------------------------------------------------------
	 out[i][0] = ux * scale
	 out[i][1] =           uy * scale
	 out[i][2] =                     uz * scale
  --------------------------------------------------------------*/
void gl_3dnow_rescale_normals_raw ( const GLmatrix *mat,
				  GLfloat scale,
				  const GLvector3f *in,
				  const GLfloat *lengths,
				  const GLubyte mask[],
				  GLvector3f *dest )
{
   (void) mat;
   (void) lengths;
   (void) mask;

   dest->count = in->count;

__asm__ (
".balign       32                                                           \n"
"                                                                           \n"
"    cmp        $0,         %3                                              \n"
"    je         4f                                                          \n"
"                                                                           \n"
"    femms                                                                  \n"
"                                                                           \n"
"    movd       %1,         %%mm0      # scale                              \n"
"    punpckldq  %%mm0,      %%mm0                                           \n"
"                                                                           \n"
"    movq       (%2),       %%mm1      # x1            | x0                 \n"
"    movd       8(%2),      %%mm2      #               | x2                 \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:  pfmul      %%mm0,      %%mm1      # x1*scale      | x0*scale           \n"
"    addl       %4,         %2         # next normal                        \n"
"                                                                           \n"
"    pfmul      %%mm0,      %%mm2      #               | x2*scale           \n"
"    addl       $12,        %0         # next r                             \n"
"                                                                           \n"
"    decl       %3                     # decrement normal counter           \n"
"    movq       %%mm1,      -12(%0)    # write r0, r1                       \n"
"                                                                           \n"
"    movd       %%mm2,      -4(%0)     # write r2                           \n"
"    movq       (%2),       %%mm1      # x1            | x0                 \n"
"                                                                           \n"
"    movd       8(%2),      %%mm2      #               | x2                 \n"
"    ja         2b                     # cnt > 0 ? -> process next normal   \n"
"                                                                           \n"
"    femms                                                                  \n"
"4:                                                                         \n"
"                                                                             "
::"a" (dest->start),
"g" (scale),
"c" (in->start),
"d" (((int) in->count)),
"g" (((int) in->stride))
:"memory");
}


