/* $Id: xform_3dnow_masked1.c,v 1.2 1999/04/16 01:14:32 brianp Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
 * 3DNow! optimizations contributed by
 * Holger Waechtler <holger@akaflieg.extern.tu-berlin.de>
 */


#include "common_3dnow.h"
#include "types.h"

/*----------------------------------------------------------------------
         ptr to source vector             v in %%eax
         ptr to matrix                    m in %%ecx
         ptr to dst vector                r in %%edx
         number of vectors to transform cnt in %%esi
         stride                      stride in %%edi
  ----------------------------------------------------------------------
     r0 = v0 * m00                       +      m30
     r1 = v0 * m01                       +      m31
     r2 = v0 * m02                       +      m32
     r3 = v0 * m03                       +      m33
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points1_general_masked ( GLvector4f *to_vec,
                        const GLmatrix *mat,
                        const GLvector4f *from_vec,
                        const GLubyte *mask,
                        const GLubyte flag )
{
to_vec->size = 4;
to_vec->flags |= VEC_SIZE_4;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
"    movq       (%%ecx),    %%mm0      # m01           | m00                \n"
"    movq       8(%%ecx),   %%mm1      # m03           | m02                \n"
"                                                                           \n"
"    movq       48(%%ecx),  %%mm2      # m31           | m30                \n"
"    movq       56(%%ecx),  %%mm3      # m33           | m32                \n"
"                                                                           \n"
"    cmp        $0,         %%esi                                           \n"
"    je         4f                                                          \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movd       (%%eax),    %%mm4      #               | x0                 \n"
"    punpckldq  %%mm4,      %%mm4      # x0            | x0                 \n"
"                                                                           \n"
"    movq       %%mm4,      %%mm5      # x0            | x0                 \n"
"    pfmul      %%mm0,      %%mm4      # x0*m01        | x0*m00             \n"
"                                                                           \n"
"    pfmul      %%mm1,      %%mm5      # x0*m03        | x0*m02             \n"
"    pfadd      %%mm2,      %%mm4      # x0*m01+m31    | x0*m00+m30         \n"
"                                                                           \n"
"    pfadd      %%mm3,      %%mm5      # x0*m03+m33    | x0*m02+m32         \n"
"    movq       %%mm4,      (%%edx)    # write r1, r0                       \n"
"                                                                           \n"
"    movq       %%mm5,      8(%%edx)   # write r3, r2                       \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    addl       $16,        %%edx      # next r                             \n"
"                                                                           \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"    ja        2b                      # cnt > 0 ? -> process next vertex   \n"
"4:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}



/*----------------------------------------------------------------------
     r0 = v0
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points1_identity_masked( GLvector4f *to_vec,
                const GLmatrix *mat,
                const GLvector4f *from_vec,
                const GLubyte *mask,
                const GLubyte flag )
{
to_vec->size = 1;
to_vec->flags |= VEC_SIZE_1;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movd       (%%eax),    %%mm0      #               | x0                 \n"
"    movd       %%mm0,      (%%edx)    #               | r0                 \n"
"                                                                           \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"    addl       $16,        %%edx      # next r                             \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"                                                                           \n"
"    ja         2b                     # cnt > 0 ? -> process next vertex   \n"
"4:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}




/*----------------------------------------------------------------------
     r0 = v0 * m00                       +      m30
     r1 = v0 * m01                       +      m31
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points1_2d_masked ( GLvector4f *to_vec,
                        const GLmatrix *mat,
                        const GLvector4f *from_vec,
                        const GLubyte *mask,
                        const GLubyte flag )
{
to_vec->size = 2;
to_vec->flags |= VEC_SIZE_2;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
"    movq       (%%ecx),    %%mm0      # m01           | m00                \n"
"    movq       48(%%ecx),  %%mm2      # m31           | m30                \n"
"                                                                           \n"
"    cmp       $0,          %%esi                                           \n"
"    je        4f                                                           \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movd       (%%eax),    %%mm4      #               | x0                 \n"
"    punpckldq  %%mm4,      %%mm4      # x0            | x0                 \n"
"                                                                           \n"
"    pfmul      %%mm0,      %%mm4      # x0*m01        | x0*m00             \n"
"    pfadd      %%mm2,      %%mm4      # x0*m01+m31    | x0*m00+m30         \n"
"                                                                           \n"
"    movq       %%mm4,      (%%edx)    # write r1, r0                       \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"                                                                           \n"
"    addl       $16,        %%edx      # next r                             \n"
"    incl       %%ebp                  # next clipmask                      \n"
"                                                                           \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"    ja        2b                      # cnt > 0 ? -> process next vertex   \n"
"4:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}




/*----------------------------------------------------------------------
     r0 = v0 * m00                       +      m30
     r1 =                                +      m31
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points1_2d_no_rot_masked ( GLvector4f *to_vec,
                        const GLmatrix *mat,
                        const GLvector4f *from_vec,
                        const GLubyte *mask,
                        const GLubyte flag )
{
to_vec->size = 2;
to_vec->flags |= VEC_SIZE_2;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
"    movd       (%%ecx),    %%mm0      #               | m00                \n"
"    movq       48(%%ecx),  %%mm2      # m31           | m30                \n"
"                                                                           \n"
"    cmp       $0,          %%esi                                           \n"
"    je        4f                                                           \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movd       (%%eax),    %%mm4      #               | x0                 \n"
"    pfmul      %%mm0,      %%mm4      #               | x0*m00             \n"
"                                                                           \n"
"    pfadd      %%mm2,      %%mm4      # m31           | x0*m00+m30         \n"
"    movq       %%mm4,      (%%edx)    # write r1, r0                       \n"
"                                                                           \n"
"3:  addl       $16,        %%edx      # next r                             \n"
"    addl       %%edi,      %%eax      # next vertex                        \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"                                                                           \n"
"    ja        2b                      # cnt > 0 ? -> process next vertex   \n"
"4:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}



/*----------------------------------------------------------------------
     r0 = v0 * m00                       +      m30
     r1 = v0 * m01                       +      m31
     r2 = v0 * m02                       +      m32
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points1_3d_masked ( GLvector4f *to_vec,
                        const GLmatrix *mat,
                        const GLvector4f *from_vec,
                        const GLubyte *mask,
                        const GLubyte flag )
{
to_vec->size = 3;
to_vec->flags |= VEC_SIZE_3;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
"    movq       (%%ecx),    %%mm0      # m01           | m00                \n"
"    movd       8(%%ecx),   %%mm1      #               | m02                \n"
"                                                                           \n"
"    movq       48(%%ecx),  %%mm2      # m31           | m30                \n"
"    movd       56(%%ecx),  %%mm3      #               | m32                \n"
"                                                                           \n"
"    cmp       $0,          %%esi                                           \n"
"    je        4f                                                           \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movd       (%%eax),    %%mm4      #               | x0                 \n"
"    punpckldq  %%mm4,      %%mm4      # x0            | x0                 \n"
"                                                                           \n"
"    movq       %%mm4,      %%mm5      #               | x0                 \n"
"    pfmul      %%mm0,      %%mm4      # x0*m01        | x0*m00             \n"
"                                                                           \n"
"    pfmul      %%mm1,      %%mm5      #               | x0*m02             \n"
"    pfadd      %%mm2,      %%mm4      # x0*m01+m31    | x0*m00+m30         \n"
"                                                                           \n"
"    pfadd      %%mm3,      %%mm5      #               | x0*m02+m32         \n"
"    movq       %%mm4,      (%%edx)    # write r1, r0                       \n"
"                                                                           \n"
"    movd       %%mm5,      8(%%edx)   # write r2                           \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"                                                                           \n"
"    addl       $16,        %%edx      # next r                             \n"
"    incl       %%ebp                  # next clipmask                      \n"
"                                                                           \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"    ja        2b                      # cnt > 0 ? -> process next vertex   \n"
"4:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}



/*----------------------------------------------------------------------
     r0 = v0 * m00                       +      m30
     r1 =                                +      m31
     r2 =                                +      m32
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points1_3d_no_rot_masked ( GLvector4f *to_vec,
                        const GLmatrix *mat,
                        const GLvector4f *from_vec,
                        const GLubyte *mask,
                        const GLubyte flag )
{
to_vec->size = 3;
to_vec->flags |= VEC_SIZE_3;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
"    movd       (%%ecx),    %%mm0      #               | m00                \n"
"    movq       48(%%ecx),  %%mm2      # m31           | m30                \n"
"                                                                           \n"
"    movd       56(%%ecx),  %%mm3      #               | m32                \n"
"    cmp        $0,         %%esi                                           \n"
"    je         4f                                                          \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movd       (%%eax),    %%mm4      #               | x0                 \n"
"    pfmul      %%mm0,      %%mm4      #               | x0*m00             \n"
"                                                                           \n"
"    pfadd      %%mm2,      %%mm4      # m31           | x0*m00+m30         \n"
"    movq       %%mm4,      (%%edx)    # write r1, r0                       \n"
"                                                                           \n"
"    movd       %%mm3,      8(%%edx)   # write r2                           \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"                                                                           \n"
"    addl       $16,        %%edx      # next r                             \n"
"    incl       %%ebp                  # next clipmask                      \n"
"                                                                           \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"    ja        2b                      # cnt > 0 ? -> process next vertex   \n"
"4:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}




/*----------------------------------------------------------------------
     r0 = v0 * m00                                  
     r1 =                     0                      
     r2 =                                       m32
     r3 =                                                   0
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points1_perspective_masked ( GLvector4f *to_vec,
                        const GLmatrix *mat,
                        const GLvector4f *from_vec,
                        const GLubyte *mask,
                        const GLubyte flag )
{
to_vec->size = 4;
to_vec->flags |= VEC_SIZE_4;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
"    movd       (%%ecx),    %%mm0      #               | m00                \n"
"    movd       56(%%ecx),  %%mm3      #               | m32                \n"
"                                                                           \n"
"    cmp        $0,         %%esi                                           \n"
"    je         4f                                                          \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movd       (%%eax),    %%mm4      # 0             | x0                 \n"
"    pfmul      %%mm0,      %%mm4      # 0             | x0*m00             \n"
"                                                                           \n"
"    movq       %%mm4,      (%%edx)    # write r1, r0                       \n"
"    movq       %%mm3,      8(%%edx)   # write r2  (=m32), r3 (=0)          \n"
"                                                                           \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"    addl       $16,        %%edx      # next r                             \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"                                                                           \n"
"    ja        2b                      # cnt > 0 ? -> process next vertex   \n"
"4:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}


