/* $Id: xform_3dnow_masked3.c,v 1.2 1999/04/16 01:14:32 brianp Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
 * 3DNow! optimizations contributed by
 * Holger Waechtler <holger@akaflieg.extern.tu-berlin.de>
 */


#include "common_3dnow.h"
#include "types.h"

/*----------------------------------------------------------------------
     r0 = v0 * m00 + v1 * m10 + v2 * m20 +      m30
     r1 = v0 * m01 + v1 * m11 + v2 * m21 +      m31
     r2 = v0 * m02 + v1 * m12 + v2 * m22 +      m32 
     r3 = v0 * m03 + v1 * m13 + v2 * m23 +      m33 
---------------------------------------------------------------------- */
void gl_3dnow_transform_points3_general_masked ( GLvector4f *to_vec,
		const GLmatrix *mat,
		const GLvector4f *from_vec,
		const GLubyte *mask,
		const GLubyte flag )
{
to_vec->size = 4;
to_vec->flags |= VEC_SIZE_4;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
"    movd       (%%ecx),    %%mm0      #               | m00                \n"
"    movd       16(%%ecx),  %%mm7      #               | m10                \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m10           |                    \n"
"    por        %%mm7,      %%mm0      # m10           | m00                \n"
"                                                                           \n"
"    movd       4(%%ecx),   %%mm1      #               | m01                \n"
"    movd       20(%%ecx),  %%mm7      #               | m11                \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m11           |                    \n"
"    por        %%mm7,      %%mm1      # m11           | m01                \n"
"                                                                           \n"
"    movq       32(%%ecx),  %%mm2      # m21           | m20                \n"
"    movq       48(%%ecx),  %%mm3      # m31           | m30                \n"
"                                                                           \n"
"    cmp        $0,         %%esi                                           \n"
"    je         6f                                                          \n"
"                                                                           \n"
"    pushl      %%ebp                                                       \n"
"    pushl      %%eax                                                       \n"
"    pushl      %%edx                                                       \n"
"    pushl      %%esi                                                       \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movq       (%%eax),    %%mm4      # x1            | x0                 \n"
"    movd       8(%%eax),   %%mm6      #               | x2                 \n"
"                                                                           \n"
"    movq       %%mm4,      %%mm5      # x1            | x0                 \n"
"    pfmul      %%mm0,      %%mm4      # x1*m10        | x0*m00             \n"
"                                                                           \n"
"    punpckldq  %%mm6,      %%mm6      # x2            | x2                 \n"
"    pfmul      %%mm1,      %%mm5      # x1*m11        | x0*m01             \n"
"                                                                           \n"
"    pfmul      %%mm2,      %%mm6      # x2*m21        | x2*m20             \n"
"    pfacc      %%mm5,      %%mm4      # x0*m01+x1*m11 | x0*m00+x1*m10      \n"
"                                                                           \n"
"    pfadd      %%mm3,      %%mm6      # x2*m21+m31    | x2*m20+m30         \n"
"    pfadd      %%mm4,      %%mm6      # r1            | r0                 \n"
"                                                                           \n"
"    movq       %%mm6,      (%%edx)    # write r0, r1                       \n"
"                                                                           \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"    addl       $16,        %%edx      # next r                             \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"                                                                           \n"
"    ja         2b                     # cnt > 0 ? -> process next vertex   \n"
"                                      # and now the second stripe ...      \n"
"                                                                           \n"
"    popl       %%esi                  # reset counter & pointers           \n"
"    popl       %%edx                                                       \n"
"    popl       %%eax                                                       \n"
"    popl       %%ebp                                                       \n"
"                                                                           \n"
"    movd       8(%%ecx),   %%mm0      #               | m02                \n"
"    movd       24(%%ecx),  %%mm7      #               | m12                \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m12           |                    \n"
"    por        %%mm7,      %%mm0      # m12           | m02                \n"
"                                                                           \n"
"    movd       12(%%ecx),  %%mm1      #               | m03                \n"
"    movd       28(%%ecx),  %%mm7      #               | m13                \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m13           |                    \n"
"    por        %%mm7,      %%mm1      # m13           | m03                \n"
"                                                                           \n"
"    movq       40(%%ecx),  %%mm2      # m23           | m22                \n"
"    movq       56(%%ecx),  %%mm3      # m33           | m32                \n"
"                                                                           \n"
".balign 32                                                                 \n"
"4:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        5f                     # skip vertex                        \n"
"                                                                           \n"
"    movq       (%%eax),    %%mm4      # x1            | x0                 \n"
"    movd       8(%%eax),   %%mm6      #               | x2                 \n"
"                                                                           \n"
"    movq       %%mm4,      %%mm5      # x1            | x0                 \n"
"    pfmul      %%mm0,      %%mm4      # x1*m12        | x0*m02             \n"
"                                                                           \n"
"    punpckldq  %%mm6,      %%mm6      # x2            | x2                 \n"
"    pfmul      %%mm1,      %%mm5      # x1*m13        | x0*m03             \n"
"                                                                           \n"
"    pfmul      %%mm2,      %%mm6      # x2*m23        | x2*m22             \n"
"    pfacc      %%mm5,      %%mm4      # x0*m03+x1*m13 | x0*m02+x1*m12      \n"
"                                                                           \n"
"    pfadd      %%mm3,      %%mm6      # x2*m23+m33    | x2*m22+m32         \n"
"    pfadd      %%mm4,      %%mm6      # r3            | r2                 \n"
"                                                                           \n"
"    movq       %%mm6,      8(%%edx)   # write r2, r3                       \n"
"                                                                           \n"
"5:  addl       %%edi,      %%eax      # next vertex                        \n"
"    addl       $16,        %%edx      # next r                             \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"                                                                           \n"
"    ja         4b                     # cnt > 0 ? -> process next vertex   \n"
"6:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}




/*----------------------------------------------------------------------
     r0 = v0
     r1 = v1
     r2 = v2
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points3_identity_masked( GLvector4f *to_vec,
                const GLmatrix *mat,
                const GLvector4f *from_vec,
                const GLubyte *mask,
                const GLubyte flag )
{
to_vec->size = 3;
to_vec->flags |= VEC_SIZE_3;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movq       (%%eax),    %%mm0      # x1            | x0                 \n"
"    movd       8(%%eax),   %%mm1      #               | x2                 \n"
"                                                                           \n"
"    movq       %%mm0,      (%%edx)    # r1            | r0                 \n"
"    movd       %%mm1,      8(%%edx)   #               | r2                 \n"
"                                                                           \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"    addl       $16,        %%edx      # next r                             \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"                                                                           \n"
"    ja         2b                     # cnt > 0 ? -> process next vertex   \n"
"4:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}




/*----------------------------------------------------------------------
     r0 = v0 * m00 + v1 * m10            +      m30
     r1 = v0 * m01 + v1 * m11            +      m31
     r2 =                       v2                  
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points3_2d_masked ( GLvector4f *to_vec,
		const GLmatrix *mat,
		const GLvector4f *from_vec,
		const GLubyte *mask,
		const GLubyte flag )
{
to_vec->size = 3;
to_vec->flags |= VEC_SIZE_3;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
"    movd       (%%ecx),    %%mm0      #               | m00                \n"
"    movd       16(%%ecx),  %%mm7      #               | m10                \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m10           |                    \n"
"    por        %%mm7,      %%mm0      # m10           | m00                \n"
"                                                                           \n"
"    movd       4(%%ecx),   %%mm1      #               | m01                \n"
"    movd       20(%%ecx),  %%mm7      #               | m11                \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m11           |                    \n"
"    por        %%mm7,      %%mm1      # m11           | m01                \n"
"                                                                           \n"
"    movq       48(%%ecx),  %%mm2      # m31           | m30                \n"
"    cmp        $0,         %%esi                                           \n"
"    je         4f                                                          \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movq       (%%eax),    %%mm3      # x1            | x0                 \n"
"    movq       %%mm3,      %%mm4      # x1            | x0                 \n"
"                                                                           \n"
"    movd       8(%%eax),   %%mm5      #               | x2                 \n"
"    pfmul      %%mm0,      %%mm3      # x1*m10        | x0*m00             \n"
"                                                                           \n"
"    pfmul      %%mm1,      %%mm4      # x1*m11        | x0*m01             \n"
"    pfacc      %%mm4,      %%mm3      # x0*m00+x1*m10 | x0*m01+x1*m11      \n"
"                                                                           \n"
"    pfadd      %%mm2,      %%mm3      # x0*...*m10+m30| x0*m01+x1*m11+m31  \n"
"    movq       %%mm3,      (%%edx)    # write r0, r1                       \n"
"                                                                           \n"
"    movd       %%mm5,      8(%%edx)   # write r2 (=x2)                     \n"
"                                                                           \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"    addl       $16,        %%edx      # next r                             \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"                                                                           \n"
"    ja         2b                     # cnt > 0 ? -> process next vertex   \n"
"4:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}




/*----------------------------------------------------------------------
     r0 = v0 * m00                       +      m30
     r1 =            v1 * m11            +      m31
     r2 =                       v2                  
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points3_2d_no_rot_masked ( GLvector4f *to_vec,
		const GLmatrix *mat,
		const GLvector4f *from_vec,
		const GLubyte *mask,
		const GLubyte flag )
{
to_vec->size = 3;
to_vec->flags |= VEC_SIZE_3;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
"    movd       (%%ecx),    %%mm0      #               | m00                \n"
"    movd       20(%%ecx),  %%mm7      #               | m11                \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m11           |                    \n"
"    por        %%mm7,      %%mm0      # m11           | m00                \n"
"                                                                           \n"
"    movq       48(%%ecx),  %%mm1      # m31           | m30                \n"
"    cmp        $0,         %%esi                                           \n"
"    je         4f                                                          \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movq       (%%eax),    %%mm4      # x1            | x0                 \n"
"    movd       8(%%eax),   %%mm5      #               | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm0,      %%mm4      # x1*m11        | x0*m00             \n"
"    movd       %%mm5,      8(%%edx)   # write r2 (=x2)                     \n"
"                                                                           \n"
"    pfadd      %%mm1,      %%mm4      # x1*m11+m31    | x0*m00+m30         \n"
"    movq       %%mm4,      (%%edx)    # write r0, r1                       \n"
"                                                                           \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"    addl       $16,        %%edx      # next r                             \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"                                                                           \n"
"    ja         2b                     # cnt > 0 ? -> process next vertex   \n"
"4:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}






/*----------------------------------------------------------------------
     r0 = v0 * m00 + v1 * m10 + v2 * m20 +      m30
     r1 = v0 * m01 + v1 * m11 + v2 * m21 +      m31
     r2 = v0 * m02 + v1 * m12 + v2 * m22 +      m32 
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points3_3d_masked ( GLvector4f *to_vec,
		const GLmatrix *mat,
		const GLvector4f *from_vec,
		const GLubyte *mask,
		const GLubyte flag )
{
to_vec->size = 3;
to_vec->flags |= VEC_SIZE_3;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
"    movd       (%%ecx),    %%mm0      #               | m00                \n"
"    movd       16(%%ecx),  %%mm7      #               | m10                \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m10           |                    \n"
"    por        %%mm7,      %%mm0      # m10           | m00                \n"
"                                                                           \n"
"    movd       4(%%ecx),   %%mm1      #               | m01                \n"
"    movd       20(%%ecx),  %%mm7      #               | m11                \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m11           |                    \n"
"    por        %%mm7,      %%mm1      # m11           | m01                \n"
"                                                                           \n"
"    movq       32(%%ecx),  %%mm2      # m21           | m20                \n"
"    movq       48(%%ecx),  %%mm3      # m31           | m30                \n"
"    cmp        $0,         %%esi                                           \n"
"    je         6f                                                          \n"
"                                                                           \n"
"    pushl      %%ebp                                                       \n"
"    pushl      %%eax                                                       \n"
"    pushl      %%edx                                                       \n"
"    pushl      %%esi                                                       \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movq       (%%eax),    %%mm4      # x1            | x0                 \n"
"    movq       %%mm4,      %%mm5      # x1            | x0                 \n"
"                                                                           \n"
"    pfmul      %%mm0,      %%mm4      # x1*m10        | x0*m00             \n"
"    movd       8(%%eax),   %%mm6      #               | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm1,      %%mm5      # x1*m11        | x0*m01             \n"
"    punpckldq  %%mm6,      %%mm6      # x2            | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm2,      %%mm6      # x2*m21        | x2*m20             \n"
"    pfacc      %%mm5,      %%mm4      # x0*m01+x1*m11 | x0*m00+x1*m10      \n"
"                                                                           \n"
"    pfadd      %%mm3,      %%mm6      # x2*m21+m31    | x2*m20+m30         \n"
"    pfadd      %%mm4,      %%mm6      # r1            | r0                 \n"
"                                                                           \n"
"    movq       %%mm6,      (%%edx)    # write r0, r1                       \n"
"                                                                           \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"    addl       $16,        %%edx      # next r                             \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"                                                                           \n"
"    ja         2b                     # cnt > 0 ? -> process next vertex   \n"
"                                      # and now the second stripe ...      \n"
"    movd       8(%%ecx),   %%mm0      #               | m02                \n"
"                                                                           \n"
"    movd       24(%%ecx),  %%mm7      #               | m12                \n"
"    psllq      $32,        %%mm7      # m12           |                    \n"
"                                                                           \n"
"    por        %%mm7,      %%mm0      # m12           | m02                \n"
"    movd       40(%%ecx),  %%mm2      #               | m22                \n"
"                                                                           \n"
"    movd       56(%%ecx),  %%mm3      #               | m32                \n"
"    popl       %%esi                  # reset counter & pointers           \n"
"                                                                           \n"
"    popl       %%edx                                                       \n"
"    popl       %%eax                                                       \n"
"                                                                           \n"
"    popl       %%ebp                                                       \n"
"                                                                           \n"
".balign 32                                                                 \n"
"4:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        5f                     # skip vertex                        \n"
"                                                                           \n"
"    movq       (%%eax),    %%mm4      # x1            | x0                 \n"
"    movq       %%mm4,      %%mm5      # x1            | x0                 \n"
"                                                                           \n"
"    pfmul      %%mm0,      %%mm4      # x1*m12        | x0*m02             \n"
"    movd       8(%%eax),   %%mm6      #               | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm1,      %%mm5      # x1*m13        | x0*m03             \n"
"    pfacc      %%mm5,      %%mm4      # x0*m03+x1*m13 | x0*m02+x1*m12      \n"
"                                                                           \n"
"    pfmul      %%mm2,      %%mm6      #               | x2*m22             \n"
"    pfadd      %%mm3,      %%mm6      #               | x2*m22+m32         \n"
"                                                                           \n"
"    pfadd      %%mm4,      %%mm6      #               | r2                 \n"
"    movd       %%mm6,      8(%%edx)   # write r2                           \n"
"                                                                           \n"
"5:  addl       %%edi,      %%eax      # next vertex                        \n"
"    addl       $16,        %%edx      # next r                             \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"                                                                           \n"
"    ja        4b                      # cnt > 0 ? -> process next vertex   \n"
"6:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}





/*----------------------------------------------------------------------
     r0 = v0 * m00                       +      m30
     r1 =            v1 * m11            +      m31
     r2 =                       v2 * m22 +      m32
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points3_3d_no_rot_masked ( GLvector4f *to_vec,
                        const GLmatrix *mat,
                        const GLvector4f *from_vec,
			const GLubyte *mask,
			const GLubyte flag )
{
to_vec->size = 3;
to_vec->flags |= VEC_SIZE_3;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
"    movd       (%%ecx),    %%mm0      #               | m00                \n"
"    movd       20(%%ecx),  %%mm7      #               | m11                \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m11           |                    \n"
"    por        %%mm7,      %%mm0      # m11           | m00                \n"
"                                                                           \n"
"    movd       40(%%ecx),  %%mm2      #               | m22                \n"
"    punpckldq  %%mm2,      %%mm2      # m22           | m22                \n"
"                                                                           \n"
"    movq       48(%%ecx),  %%mm1      # m31           | m30                \n"
"    movd       56(%%ecx),  %%mm3      #               | m32                \n"
"                                                                           \n"
"    punpckldq  %%mm3,      %%mm3      # m32           | m32                \n"
"    cmp        $0,         %%esi                                           \n"
"    je         4f                                                          \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movq       (%%eax),    %%mm4      # x1            | x0                 \n"
"    movd       8(%%eax),   %%mm5      #               | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm0,      %%mm4      # x1*m11        | x0*m00             \n"
"    pfmul      %%mm2,      %%mm5      #               | x2*m22             \n"
"                                                                           \n"
"    pfadd      %%mm1,      %%mm4      # x1*m11+m31    | x0*m00+m30         \n"
"    pfadd      %%mm3,      %%mm5      #               | x2*m22+m32         \n"
"                                                                           \n"
"    movq       %%mm4,      (%%edx)    # write r0, r1                       \n"
"    movd       %%mm5,      8(%%edx)   # write r2                           \n"
"                                                                           \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"    addl       $16,        %%edx      # next r                             \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"                                                                           \n"
"    ja         2b                     # cnt > 0 ? -> process next vertex   \n"
"4:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}




/*----------------------------------------------------------------------
     r0 = v0 * m00            + v2 * m20              
     r1 =            v1 * m11 + v2 * m21           
     r2 =                       v2 * m22 +      m32
     r3 =                     - v2                  
  ---------------------------------------------------------------------- */
void gl_3dnow_transform_points3_perspective_masked ( GLvector4f *to_vec,
		const GLmatrix *mat,
		const GLvector4f *from_vec,
		const GLubyte *mask,
		const GLubyte flag )
{
to_vec->size = 4;
to_vec->flags |= VEC_SIZE_4;
to_vec->count = from_vec->count;
BEGIN_XFORM_3DNOW_ASM_MASKED
"                                                                           \n"
"    movd       (%%ecx),    %%mm0      #               | m00                \n"
"    movd       20(%%ecx),  %%mm7      #               | m11                \n"
"                                                                           \n"
"    psllq      $32,        %%mm7      # m11           |                    \n"
"    por        %%mm7,      %%mm0      # m11           | m00                \n"
"                                                                           \n"
"    movq       32(%%ecx),  %%mm1      # m21           | m20                \n"
"    movd       40(%%ecx),  %%mm2      #               | m22                \n"
"                                                                           \n"
"    movd       56(%%ecx),  %%mm3      #               | m32                \n"
"    cmp        $0,         %%esi                                           \n"
"    je         4f                                                          \n"
"                                                                           \n"
".balign 32                                                                 \n"
"2:                                                                         \n"
"    testb      %%bl,       (%%ebp)    # mask [i] != clip flag ??           \n"
"    jnz        3f                     # skip vertex                        \n"
"                                                                           \n"
"    movq       (%%eax),    %%mm4      # x1            | x0                 \n"
"    movd       8(%%eax),   %%mm5      #               | x2                 \n"
"                                                                           \n"
"    pfmul      %%mm0,      %%mm4      # x1*m11        | x0*m00             \n"
"    movq       %%mm5,      %%mm6      #               | x2                 \n"
"                                                                           \n"
"    punpckldq  %%mm5,      %%mm5      # x2            | x2                 \n"
"    pfmul      %%mm1,      %%mm5      # x2*m21        | x2*m20             \n"
"                                                                           \n"
"    pfadd      %%mm4,      %%mm5      # x1*m11+x2*m21 | x0*m00+x2*m20      \n"
"    movq       %%mm5,      (%%edx)    # write r0, r1                       \n"
"                                                                           \n"
"    movq       %%mm6,      %%mm5      #               | x2                 \n"
"    pfmul      %%mm2,      %%mm5      #               | x2*m22             \n"
"                                                                           \n"
"    pfadd      %%mm3,      %%mm5      #               | x2*m22+m32         \n"
"    pfsubr     %%mm7,      %%mm6      # (LO mm7 == 0) | -x2                \n"
"                                                                           \n"
"    movd       %%mm5,      8(%%edx)   # write r2                           \n"
"    movd       %%mm6,      12(%%edx)  # write r3                           \n"
"                                                                           \n"
"3:  addl       %%edi,      %%eax      # next vertex                        \n"
"    addl       $16,        %%edx      # next r                             \n"
"                                                                           \n"
"    incl       %%ebp                  # next clipmask                      \n"
"    decl       %%esi                  # decrement vertex counter           \n"
"                                                                           \n"
"    ja         2b                     # cnt > 0 ? -> process next vertex   \n"
"4:                                                                         \n"
END_XFORM_3DNOW_ASM_MASKED
}

