/* -*- mode: C; tab-width:8;  -*-

             fxcva.c - the CVA related code
*/

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * See the file fxapi.c for more informations about authors
 *
 */


#if defined(FX)

#include "fxdrv.h"
#include "mmath.h"
#include "vbindirect.h"
#include "pb.h"
#include "fxvsetup.h"

#define PRIM_POINTS 0
#define PRIM_LINES  1
#define PRIM_TRIS   2
#define PRIM_CULLED 3

static void kgrDrawTriangle(GrVertex *v1, GrVertex *v2, GrVertex *v3)
{}

/*  #define grDrawTriangle kgrDrawTriangle */


static GLuint reduce_prim[GL_POLYGON+2] = {
   PRIM_POINTS,
   PRIM_LINES,
   PRIM_LINES,
   PRIM_LINES,
   PRIM_TRIS,
   PRIM_TRIS,
   PRIM_TRIS,
   PRIM_TRIS,
   PRIM_TRIS,
   PRIM_TRIS,
   PRIM_CULLED
};

typedef void (*mergefunc)( struct vertex_buffer *cvaVB,
			   struct vertex_buffer *VB,
			   const struct gl_prim_state *state,
			   GLuint start,
			   GLuint count );

static void fxCvaRenderNoop( struct vertex_buffer *cvaVB,
			     struct vertex_buffer *VB,
			     const struct gl_prim_state *state,
			     GLuint start,
			     GLuint count )
{
}

static INLINE void fxRenderClippedTriangle2( struct vertex_buffer *VB,
					     GLuint v1, GLuint v2, GLuint v3 )
{
  GrVertex *gWin=(GrVertex *)VB->driver_data;
  GLubyte *clipmask = VB->ClipMask;
  GLubyte mask = clipmask[v1] | clipmask[v2] | clipmask[v3];

  if (!mask) 
    grDrawTriangle(&gWin[v1], &gWin[v2], &gWin[v3]);
  else if (!(clipmask[v1]&clipmask[v2]&clipmask[v3]&CLIP_ALL_BITS)) 
  {    
    GLuint n;
    GLuint vlist[VB_MAX_CLIPPED_VERTS];
    ASSIGN_3V(vlist, v1, v2, v3);

    n = (gl_poly_clip_tab[VB->ClipPtr->size])( VB, 3, vlist, mask );
    if (n >= 3) {
      GLuint i, j0 = vlist[0];
      for (i=2;i<n;i++) 
	grDrawTriangle(&gWin[j0],&gWin[vlist[i-1]],&gWin[vlist[i]]);
    }
  }
}


static mergefunc merge_and_render_tab[2][MAX_MERGABLE][PRIM_CULLED+1];



#define CVA_VARS_RGBA 					\
   VARS_RGBA						\
   GLubyte (*cva_color)[4] = cvaVB->ColorPtr->data;

#define CVA_VARS_TMU0							\
   VARS_TMU0								\
   GLfloat (*cva_tex0)[4] = cvaVB->TexCoordPtr[tmu0_source]->data;

#define CVA_VARS_TMU1							\
   VARS_TMU1								\
   GLfloat (*cva_tex1)[4] = cvaVB->TexCoordPtr[tmu1_source]->data;


#define INIT_RGBA (void) cva_color;
#define INIT_TMU0 (void) cva_tex0; (void) tmu0_sz;
#define INIT_TMU1 (void) cva_tex1; (void) tmu1_sz;

#define DRAW_POINT grDrawPoint( v )
#define DRAW_LINE grDrawLine( v, prev_v ) 
#define DRAW_TRI grDrawTriangle( &gWin[l[0]], &gWin[l[1]], v )
#define DRAW_TRI2 grDrawTriangle( vl[0], vl[1], v )
#define CLIP_LINE fxRenderClippedLine( cvaVB, e, prev )
#define CLIP_OR_DRAW_TRI fxRenderClippedTriangle2( cvaVB, l[0], l[1], e )
#define DIRECT 1

#define TAG(x) x
#define IDX 0
#define VARS
#define INIT
#define MERGE_RAST
#define MERGE_VB
#include "fxcvatmp.h"

#define TAG(x) x##RGBA
#define IDX SETUP_RGBA
#define VARS CVA_VARS_RGBA
#define INIT INIT_RGBA
#define MERGE_RAST DO_SETUP_RGBA
#define MERGE_VB COPY_4UBV(cva_color[e], color[i])
#include "fxcvatmp.h"

#define TAG(x) x##T0
#define IDX SETUP_TMU0
#define VARS CVA_VARS_TMU0
#define INIT INIT_TMU0
#define MERGE_RAST DO_SETUP_TMU0
#define MERGE_VB COPY_2FV(cva_tex0[e], tmu0_data[i])
#include "fxcvatmp.h"

#define TAG(x) x##T1
#define IDX SETUP_TMU1
#define VARS CVA_VARS_TMU1
#define INIT INIT_TMU1
#define MERGE_RAST DO_SETUP_TMU1
#define MERGE_VB COPY_2FV(cva_tex1[e], tmu1_data[i])
#include "fxcvatmp.h"

#define TAG(x) x##T0T1
#define IDX SETUP_TMU0|SETUP_TMU1
#define VARS CVA_VARS_TMU0 CVA_VARS_TMU1 
#define INIT INIT_TMU0 INIT_TMU1 
#define MERGE_RAST DO_SETUP_TMU0 DO_SETUP_TMU1
#define MERGE_VB COPY_2FV(cva_tex0[e], tmu0_data[i]); \
                 COPY_2FV(cva_tex1[e], tmu1_data[i]);
#include "fxcvatmp.h"

#define TAG(x) x##RGBAT0
#define IDX SETUP_RGBA|SETUP_TMU0
#define VARS CVA_VARS_RGBA CVA_VARS_TMU0 
#define INIT INIT_RGBA INIT_TMU0 
#define MERGE_RAST DO_SETUP_RGBA DO_SETUP_TMU0
#define MERGE_VB COPY_2FV(cva_tex0[e], tmu0_data[i]); \
                 COPY_4UBV(cva_color[e], color[i]);
#include "fxcvatmp.h"

#define TAG(x) x##RGBAT1
#define IDX SETUP_RGBA|SETUP_TMU1
#define VARS CVA_VARS_RGBA CVA_VARS_TMU1
#define INIT INIT_RGBA INIT_TMU1
#define MERGE_RAST DO_SETUP_RGBA DO_SETUP_TMU1
#define MERGE_VB COPY_2FV(cva_tex1[e], tmu1_data[i]); \
                 COPY_4UBV(cva_color[e], color[i]);
#include "fxcvatmp.h"

#define TAG(x) x##RGBAT0T1
#define IDX SETUP_RGBA|SETUP_TMU0|SETUP_TMU1
#define VARS CVA_VARS_RGBA CVA_VARS_TMU0 CVA_VARS_TMU1
#define INIT INIT_RGBA INIT_TMU0 INIT_TMU1
#define MERGE_RAST DO_SETUP_RGBA DO_SETUP_TMU0 DO_SETUP_TMU1
#define MERGE_VB COPY_2FV(cva_tex0[e], tmu0_data[i]); \
                 COPY_2FV(cva_tex1[e], tmu1_data[i]); \
                 COPY_4UBV(cva_color[e], color[i]);
#include "fxcvatmp.h"



#undef DRAW_POINT 
#undef DRAW_LINE 
#undef DRAW_TRI 
#undef CLIP_LINE 
#undef CLIP_OR_DRAW_TRI 
#undef DIRECT

#define DRAW_POINT ctx->Driver.PointsFunc( ctx, e, e )
#define DRAW_LINE ctx->Driver.LineFunc( ctx, e, prev, e ) 
#define DRAW_TRI ctx->TriangleFunc( ctx, l[0], l[1], e, e )
#define CLIP_LINE gl_render_clipped_line( ctx, e, prev )
#define CLIP_OR_DRAW_TRI 						\
do {									\
   if (clip[l[0]] | clip[l[1]] | clip[e]) {				\
      if (!(clip[l[0]] & clip[l[1]] & clip[e] & CLIP_ALL_BITS)) {	\
	 COPY_3V(vlist, l);						\
	 gl_render_clipped_triangle( ctx, 3, vlist, e );		\
      }									\
   }									\
   else ctx->TriangleFunc( ctx, l[0], l[1], e, e );			\
} while (0)

	    
#define DIRECT 0

#define TAG(x) x##_indirect
#define IDX 0
#define VARS
#define INIT
#define MERGE_RAST
#define MERGE_VB
#include "fxcvatmp.h"

#define TAG(x) x##RGBA_indirect
#define IDX SETUP_RGBA
#define VARS CVA_VARS_RGBA
#define INIT INIT_RGBA
#define MERGE_RAST DO_SETUP_RGBA
#define MERGE_VB COPY_4UBV(cva_color[e], color[i])
#include "fxcvatmp.h"

#define TAG(x) x##T0_indirect
#define IDX SETUP_TMU0
#define VARS CVA_VARS_TMU0
#define INIT INIT_TMU0
#define MERGE_RAST DO_SETUP_TMU0
#define MERGE_VB COPY_2FV(cva_tex0[e], tmu0_data[i])
#include "fxcvatmp.h"

#define TAG(x) x##T1_indirect
#define IDX SETUP_TMU1
#define VARS CVA_VARS_TMU1
#define INIT INIT_TMU1
#define MERGE_RAST DO_SETUP_TMU1
#define MERGE_VB COPY_2FV(cva_tex1[e], tmu1_data[i])
#include "fxcvatmp.h"

#define TAG(x) x##T0T1_indirect
#define IDX SETUP_TMU0|SETUP_TMU1
#define VARS CVA_VARS_TMU0 CVA_VARS_TMU1 
#define INIT INIT_TMU0 INIT_TMU1 
#define MERGE_RAST DO_SETUP_TMU0 DO_SETUP_TMU1
#define MERGE_VB COPY_2FV(cva_tex0[e], tmu0_data[i]); \
                 COPY_2FV(cva_tex1[e], tmu1_data[i]);
#include "fxcvatmp.h"

#define TAG(x) x##RGBAT0_indirect
#define IDX SETUP_RGBA|SETUP_TMU0
#define VARS CVA_VARS_RGBA CVA_VARS_TMU0 
#define INIT INIT_RGBA INIT_TMU0 
#define MERGE_RAST DO_SETUP_RGBA DO_SETUP_TMU0
#define MERGE_VB COPY_2FV(cva_tex0[e], tmu0_data[i]); \
                 COPY_4UBV(cva_color[e], color[i]);
#include "fxcvatmp.h"

#define TAG(x) x##RGBAT1_indirect
#define IDX SETUP_RGBA|SETUP_TMU1
#define VARS CVA_VARS_RGBA CVA_VARS_TMU1
#define INIT INIT_RGBA INIT_TMU1
#define MERGE_RAST DO_SETUP_RGBA DO_SETUP_TMU1
#define MERGE_VB COPY_2FV(cva_tex1[e], tmu1_data[i]); \
                 COPY_4UBV(cva_color[e], color[i]);
#include "fxcvatmp.h"

#define TAG(x) x##RGBAT0T1_indirect
#define IDX SETUP_RGBA|SETUP_TMU0|SETUP_TMU1
#define VARS CVA_VARS_RGBA CVA_VARS_TMU0 CVA_VARS_TMU1
#define INIT INIT_RGBA INIT_TMU0 INIT_TMU1
#define MERGE_RAST DO_SETUP_RGBA DO_SETUP_TMU0 DO_SETUP_TMU1
#define MERGE_VB COPY_2FV(cva_tex0[e], tmu0_data[i]); \
                 COPY_2FV(cva_tex1[e], tmu1_data[i]); \
                 COPY_4UBV(cva_color[e], color[i]);
#include "fxcvatmp.h"




void fxDDCvaInit()
{
   /* Call grDrawTriangle et al */
   init_cva();
   init_cvaT0();
   init_cvaT1();
   init_cvaT0T1();
   init_cvaRGBA();
   init_cvaRGBAT0();
   init_cvaRGBAT1();
   init_cvaRGBAT0T1();

   /* Call ctx->TriangleFunc and friends */
   init_cva_indirect();
   init_cvaT0_indirect();
   init_cvaT1_indirect();
   init_cvaT0T1_indirect();
   init_cvaRGBA_indirect();
   init_cvaRGBAT0_indirect();
   init_cvaRGBAT1_indirect();
   init_cvaRGBAT0T1_indirect();
}


void fxDDCheckMergeAndRender( GLcontext *ctx, struct gl_pipeline_stage *d )
{
   fxMesaContext fxMesa = (fxMesaContext)ctx->DriverCtx;
   
/*    if (fxMesa->mergeindex == 0) */
/*       return; */

   if (ctx->IndirectTriangles)
      return;

   d->inputs = (VERT_RAST_SETUP_PART|VERT_ELT|fxMesa->mergeinputs);
   d->outputs = 0;
   d->type = PIPE_CVA_ELT;
}


extern void fxPointSmooth(GLcontext *ctx, GLuint first, GLuint last);
extern void fxLineSmooth(GLcontext *ctx, GLuint v1, GLuint v2, GLuint pv);
extern void fxTriangleSmooth(GLcontext *ctx, GLuint v1, GLuint v2, GLuint v3, 
			     GLuint pv);
      extern const char *gl_prim_name[];


static GLboolean edge_flag[GL_POLYGON+2] = { 0,0,0,0,1,0,0,1,0,1,0 };

void fxDDMergeAndRender( struct vertex_buffer *VB,
			 struct vertex_buffer *cvaVB )
{
   GLcontext *ctx = VB->ctx;
   fxMesaContext fxMesa = (fxMesaContext)ctx->DriverCtx;
   GLuint i, next, prim;
   GLuint parity = VB->Parity;
   GLuint count = VB->Count;
   const struct gl_prim_state *state;
   mergefunc func;
   struct vertex_buffer *saved_vb = ctx->VB;
   GLuint inputs = ctx->RenderFlags & ~ctx->CVA.pre.outputs;
   GLuint flags = 0;

   GLuint direct = (fxMesa->render_index == 0);

   mergefunc (*tab)[PRIM_CULLED+1] = merge_and_render_tab[direct];

   /* Should be precalc'ed.  May actually contain elements not present
    * in fxMesa->setupindex, eg RGBA when flat shading.  These need
    * to be copied into the cva VB so that funcs like fxTriangleFlat
    * will be able to reach them.
    *
    * This leads to some duplication of effort in the merge funcs.
    */
   if (inputs & VERT_RGBA) 
      flags |= SETUP_RGBA;

   if (inputs & VERT_TEX0_ANY)
      if (fxMesa->tmu_source[0] == 0) 
	 flags |= SETUP_TMU0; 
      else
	 flags |= SETUP_TMU1;
      
   if (inputs & VERT_TEX1_ANY)
      if (fxMesa->tmu_source[0] == 1) 
	 flags |= SETUP_TMU0; 
      else
	 flags |= SETUP_TMU1;


   ctx->VB = cvaVB;

   for ( i= VB->CopyStart ; i < count ; parity = 0, i = next ) 
   {
      prim = VB->Primitive[i];
      next = VB->NextPrimitive[i];
      cvaVB->EdgeFlagPrimitive = edge_flag[prim];

      state = gl_prim_state_machine[prim][parity];
      func = tab[flags][reduce_prim[prim]];

/*       fxPrintSetupFlags("merge", flags); */
/*       printf("fxMergeAndRender: %s/%d %d-%d (%d)\n", gl_prim_name[prim], parity, */
/*   	     i, next, direct); */
      
      func( cvaVB, VB, state, i, next );
      
      if (ctx->LightTwoSide) {
	 cvaVB->Specular = cvaVB->Spec[0];
	 cvaVB->ColorPtr = cvaVB->Color[0];
	 cvaVB->IndexPtr = cvaVB->Index[0];
      }
   }

   if (ctx->PB->count > 0) 
      gl_flush_pb(ctx);

   ctx->VB = saved_vb;
}

extern void fxDDRenderElementsDirect( struct vertex_buffer *VB,
				      GLenum prim,
				      GLuint *elt,
				      GLuint nr );


/* Replacements for gl_render_vb_indirect and gl_render_elts which know how
 * to call grDrawTriangle directly...
 *
 * If 'direct' evaluates to false, these functions are identical to those
 * they replace.
 */
void fxDDRenderCVAElements( struct vertex_buffer *VB,
			    GLenum prim,
			    GLuint *elt,
			    GLuint nr )
{
   GLcontext *ctx = VB->ctx;
   fxMesaContext fxMesa = (fxMesaContext)ctx->DriverCtx;

   if (fxMesa->render_index == 0)
      fxDDRenderElementsDirect( VB, prim, elt, nr );
   else 
   {
      struct vertex_buffer *saved_vb = ctx->VB;
      GLuint *saved_elts = VB->Elt;
      const struct gl_prim_state *state = gl_prim_state_machine[prim][0];
      mergefunc func = merge_and_render_tab[0][0][reduce_prim[prim]];

      ctx->VB = VB;
      VB->Elt = elt;		/* kludge */

/*       printf("fxDDRenderCVAElements: %s/%d %d-%d\n", gl_prim_name[prim], 0, */
/* 	     0, nr); */

      func( VB, VB, state, 0, nr );
      
      if (ctx->LightTwoSide) {
	 VB->Specular = VB->Spec[0];
	 VB->ColorPtr = VB->Color[0];
	 VB->IndexPtr = VB->Index[0];
      }

      if (ctx->PB->count > 0) 
	 gl_flush_pb(ctx);

      VB->Elt = saved_elts;
      ctx->VB = saved_vb;
   }
}



void fxDDRenderVBIndirect( struct vertex_buffer *VB,
			   struct vertex_buffer *cvaVB )
{

   GLcontext *ctx = VB->ctx;
   GLuint i, next, prim;
   GLuint parity = VB->Parity;
   GLuint count = VB->Count;
   const struct gl_prim_state *state;
   mergefunc func;
   struct vertex_buffer *saved_vb = ctx->VB;
   fxMesaContext fxMesa = (fxMesaContext)ctx->DriverCtx;
   GLuint direct = (fxMesa->render_index == 0);

   mergefunc (*tab)[PRIM_CULLED+1] = merge_and_render_tab[direct];

   ctx->VB = cvaVB;

   for ( i= VB->CopyStart ; i < count ; parity = 0, i = next ) 
   {
      prim = VB->Primitive[i];
      next = VB->NextPrimitive[i];
      cvaVB->EdgeFlagPrimitive = edge_flag[prim];

/*       printf("fxDDRenderVBIndirect: %s/%d %d-%d\n", gl_prim_name[prim], parity, */
/* 	     i, next); */

      state = gl_prim_state_machine[prim][parity];
      func = tab[0][reduce_prim[prim]];
      
      func( cvaVB, VB, state, i, next );
      
      if (ctx->LightTwoSide) {
	 cvaVB->Specular = cvaVB->Spec[0];
	 cvaVB->ColorPtr = cvaVB->Color[0];
	 cvaVB->IndexPtr = cvaVB->Index[0];
      }
   }

   if (ctx->PB->count > 0) 
      gl_flush_pb(ctx);

   ctx->VB = saved_vb;

}

#else


/*
 * Need this to provide at least one external definition.
 */

int gl_fx_dummy_function_cva(void)
{
  return 0;
}


#endif   /* FX */
