
static void TAG(cva_render_points)( struct vertex_buffer *cvaVB,
				    struct vertex_buffer *VB,
				    const struct gl_prim_state *state,
				    GLuint start,
				    GLuint count ) 
{
   GLcontext *ctx = VB->ctx;
   fxMesaContext fxMesa = (fxMesaContext)ctx->DriverCtx;
   GrVertex *gWin = (GrVertex *) cvaVB->driver_data;
   const GLuint *elt = VB->Elt;
   GLuint i;

   VARS;
   INIT;
   (void) fxMesa;

   if (cvaVB->ClipOrMask) {
      const GLubyte *clip = cvaVB->ClipMask;
      for (i = start ; i < count ; i++) {
	 GLuint e = elt[i];
	 if (!clip[e]) {
	    GrVertex *v = &gWin[e]; (void) v;
	    if (!DIRECT) { MERGE_VB; }
	    MERGE_RAST;
	    DRAW_POINT;
	 }
      }
   } else {
      for (i = start ; i < count ; i++) {
	 GLuint e = elt[i];
	 GrVertex *v = &gWin[e]; (void) v;
	 if (!DIRECT) { MERGE_VB; }
	 MERGE_RAST;
	 DRAW_POINT;
      }
   }
}

static void TAG(cva_render_lines)( struct vertex_buffer *cvaVB,
				   struct vertex_buffer *VB,
				   const struct gl_prim_state *state,
				   GLuint start,
				   GLuint count ) 
{
   GLcontext *ctx = VB->ctx;
   fxMesaContext fxMesa = (fxMesaContext)ctx->DriverCtx;
   GrVertex *gWin = (GrVertex *) cvaVB->driver_data;
   const GLuint *elt = VB->Elt;
   GLuint i;

   VARS;
   INIT;
   (void) fxMesa;

   if (cvaVB->ClipOrMask) {
      const GLubyte *clip = cvaVB->ClipMask;
      GLuint prev = 0;
      GrVertex *prev_v = 0;

      (void) ctx;
   
      for (i = start ; i < count ; i++) {
	 GLuint e = elt[i];
	 GrVertex *v = &gWin[e];

	 MERGE_VB;

	 if (!clip[e])
	    MERGE_RAST;

	 if (state->draw)
	    if (clip[e] | clip[prev]) 
	       CLIP_LINE;
	    else
	       DRAW_LINE;

	 prev = e;	 
	 prev_v = v;
	 state = state->next;
      }
      if (state->finish_loop) {
	 GLuint e = elt[start];
	 GrVertex *v = &gWin[e]; (void) v;

	 if (!DIRECT) { MERGE_VB; }
	 MERGE_RAST;

	 if (clip[e] | clip[prev]) 
	    CLIP_LINE;
	 else
	    DRAW_LINE;
      }
   } else {
      GLuint prev = 0;
      GrVertex *prev_v = 0;
      
      for (i = start ; i < count ; i++) {
	 GLuint e = elt[i];
	 GrVertex *v = &gWin[e];

	 if (!DIRECT) { MERGE_VB; }
	 MERGE_RAST;
	 if (state->draw) DRAW_LINE;
	 prev = e;
	 prev_v = v;
	 state = state->next;
      }
      if (state->finish_loop) {
	 GLuint e = elt[start];
	 GrVertex *v = &gWin[e]; (void) v;

	 if (!DIRECT) { MERGE_VB; }
	 MERGE_RAST;
	 DRAW_LINE;
      }
   }      
}


static void TAG(cva_render_tris)( struct vertex_buffer *cvaVB,
				  struct vertex_buffer *VB,
				  const struct gl_prim_state *state,
				  GLuint start,
				  GLuint count ) 
{
   GLcontext *ctx = VB->ctx;
   fxMesaContext fxMesa = (fxMesaContext)ctx->DriverCtx;
   GrVertex *gWin = (GrVertex *) cvaVB->driver_data;
   const GLuint *elt = VB->Elt;
   GLuint i;
   VARS;
   INIT;
   (void) fxMesa;

   if (cvaVB->ClipOrMask) {      
      GLuint vlist[VB_MAX_CLIPPED_VERTS];
      GLuint l[3];
      const GLubyte *clip = cvaVB->ClipMask; 
      
      (void) vlist;

      for (i = start ; i < count ; i++) {
	 GLuint e = l[2] = elt[i];
	 GrVertex *v = &gWin[e]; (void) v;

	 MERGE_VB;		/* needed for clip-interp */

	 if (!clip[e]) {
	    MERGE_RAST;
	 }

	 if (state->draw) 
	    CLIP_OR_DRAW_TRI;


	 l[0] = l[state->v0];
	 l[1] = l[state->v1];
	 state = state->next;
      }
   } else if (DIRECT) {
      GrVertex *vl[3];
      for (i = start ; i < count ; i++) {
	 GLuint e = elt[i];
	 GrVertex *v = vl[2] = &gWin[e];

	 MERGE_RAST;

	 if (state->draw) 
	    DRAW_TRI2;

	 vl[0] = vl[state->v0];
	 vl[1] = vl[state->v1];
	 state = state->next;
      }
   } else {
      GLuint l[3];
      for (i = start ; i < count ; i++) {
	 GLuint e = l[2] = elt[i];
	 GrVertex *v = &gWin[e]; (void) v;

	 MERGE_VB;		/* needed for ctx->trianglefunc? */
	 MERGE_RAST;
	 
	 if (state->draw) 
	    DRAW_TRI;

	 l[0] = l[state->v0];
	 l[1] = l[state->v1];
	 state = state->next;
      }
   }
}


static void TAG(init_cva)( void )
{
   merge_and_render_tab[DIRECT][IDX][PRIM_POINTS] = TAG(cva_render_points);
   merge_and_render_tab[DIRECT][IDX][PRIM_LINES] = TAG(cva_render_lines);
   merge_and_render_tab[DIRECT][IDX][PRIM_TRIS] = TAG(cva_render_tris);
   merge_and_render_tab[DIRECT][IDX][PRIM_CULLED] = fxCvaRenderNoop;
}


#undef IDX
#undef MERGE_RAST
#undef MERGE_VB
#undef VARS
#undef INIT
#undef TAG

