/* -*- mode: C; tab-width:8; c-basic-offset:2 -*-

             fxrender.c - 3Dfx VooDoo RenderVB driver function support
*/

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * See the file fxapi.c for more informations about authors
 *
 */

#if defined(FX)

#include "fxdrv.h"
#include "vbcull.h"

static void kgrDrawTriangle(const GrVertex *v1, const GrVertex *v2, const GrVertex *v3)
{}

/*  #define grDrawTriangle kgrDrawTriangle */

/*
 * Render a line segment from VB[v1] to VB[v2] when either one or both
 * endpoints must be clipped.
 */
INLINE void fxRenderClippedLine( struct vertex_buffer *VB, 
				 GLuint v1, GLuint v2 )
{
  GrVertex *gWin=(GrVertex *)VB->driver_data;
  GLubyte mask = VB->ClipMask[v1] | VB->ClipMask[v2];

  if (!mask || (gl_line_clip_tab[VB->ClipPtr->size])(VB, &v1, &v2, mask))
    grDrawLine(&gWin[v1],&gWin[v2]);
}


/* This is legal for Quads as well as triangles, hence the 'n' parameter.
 */
INLINE void fxRenderClippedTriangle( struct vertex_buffer *VB,
					 GLuint n, GLuint vlist[] )
{
  GLubyte mask = 0;
  GLuint i;

  for (i = 0 ; i < n ; i++) 
    mask |= VB->ClipMask[vlist[i]];
   
  n = (gl_poly_clip_tab[VB->ClipPtr->size])( VB, n, vlist, mask );
  if (n >= 3) {
    GrVertex *gWin=(GrVertex *)VB->driver_data;
    GrVertex *i0 = &gWin[vlist[0]];
    GrVertex *i1 = &gWin[vlist[1]];
    GrVertex *i2 = &gWin[vlist[2]];
    GLuint i;
      
    for (i=2;i<n;i++, i1 = i2, i2 = &gWin[vlist[i]]) 
      grDrawTriangle(i0,i1,i2);
  }
}


static INLINE void fxRenderClippedTriangle2( struct vertex_buffer *VB,
					     GLuint v1, GLuint v2, GLuint v3 )
{
  GrVertex *gWin=(GrVertex *)VB->driver_data;
  GLubyte *clipmask = VB->ClipMask;
  GLubyte mask = clipmask[v1] | clipmask[v2] | clipmask[v3];

  if (!mask) 
    grDrawTriangle(&gWin[v1], &gWin[v2], &gWin[v3]);
  else if (!(clipmask[v1]&clipmask[v2]&clipmask[v3]&CLIP_ALL_BITS)) 
  {    
    GLuint n;
    GLuint vlist[VB_MAX_CLIPPED_VERTS];
    ASSIGN_3V(vlist, v1, v2, v3);

    n = (gl_poly_clip_tab[VB->ClipPtr->size])( VB, 3, vlist, mask );
    if (n >= 3) {
      GLuint i;
      GrVertex *i0 = &gWin[vlist[0]];
      GrVertex *i1 = &gWin[vlist[1]];
      GrVertex *i2 = &gWin[vlist[2]];
      
      for (i=2;i<n;i++, i1 = i2, i2 = &gWin[vlist[i]]) 
	grDrawTriangle(i0,i1,i2);
    }
  }
}



/************************************************************************/
/************************ RenderVB functions ****************************/
/************************************************************************/

static render_func render_vb_smooth_raw[GL_POLYGON+2];
static render_func render_vb_smooth_culled[GL_POLYGON+2];
static render_func render_vb_flat_raw[GL_POLYGON+2];
static render_func render_vb_flat_culled[GL_POLYGON+2];
static render_func render_vb_smooth_indirect_clipped[GL_POLYGON+2];
static render_func render_vb_smooth_indirect[GL_POLYGON+2];



static void render_vb_noop( struct vertex_buffer *VB,
			    GLuint start,
			    GLuint count,
			    GLuint parity )
{
   /* Handles completely culled primitives - not much to do.
    */
}


/* Render front-facing, non-clipped primitives.
 */

#define RENDER_POINTS( start, count )		\
   (void) gWin;						\
   (void) VB;						\
   (VB->ctx->Driver.PointsFunc)( VB->ctx, start, count-1 )


#define RENDER_LINE( i1, i ) 		\
do {						\
     RVB_COLOR(i);				\
     grDrawLine(&gWin[i1],&gWin[i]);       	\
} while (0)

#define RENDER_TRI( i2, i1, i, pv, parity ) 		\
do {								\
     RVB_COLOR(pv);						\
     if (parity)						\
       grDrawTriangle(&gWin[i1],&gWin[i2],&gWin[i]);		\
     else							\
       grDrawTriangle(&gWin[i2],&gWin[i1],&gWin[i]);		\
} while (0)

#define RENDER_QUAD( i3, i2, i1, i, pv )  		\
do {							\
     RVB_COLOR(pv);					\
     grDrawTriangle(&gWin[i3],&gWin[i2],&gWin[i]);	\
     grDrawTriangle(&gWin[i2],&gWin[i1],&gWin[i]);	\
} while (0)



#define LOCAL_VARS					\
   fxMesaContext fxMesa=(fxMesaContext)VB->ctx->DriverCtx;	\
   const GrVertex *gWin=(GrVertex *)VB->driver_data;	\
   (void) fxMesa;
#define INIT(x)  

#define TAB render_vb_flat_raw
#define TAG(x) x##_fx_flat_raw
#undef  RVB_COLOR
#define RVB_COLOR(pv)  FX_VB_COLOR(fxMesa, VB->ColorPtr->data[pv])
#define PRESERVE_VB_DEFS

#include "render_tmp.h"

#define TAB    render_vb_smooth_raw
#define TAG(x) x##_fx_smooth_raw
#undef  RVB_COLOR
#define RVB_COLOR(x)

#include "render_tmp.h"



/* Render with clipped and/or culled primitives with cullmask information.
 */
#define RENDER_POINTS( start, count )		\
   (void) gWin;						\
   (void) cullmask;					\
   (VB->ctx->Driver.PointsFunc)( VB->ctx, start, count-1 )


#define RENDER_LINE( i1, i )		\
do {						\
   const GLubyte flags = cullmask[i];		\
						\
   if (!(flags & PRIM_NOT_CULLED))		\
      continue;					\
						\
   RVB_COLOR(i);				\
   if (flags & PRIM_ANY_CLIP)			\
      fxRenderClippedLine( VB, i1, i );	\
   else						\
      grDrawLine( &gWin[i1], &gWin[i] );	\
} while (0)


#define RENDER_TRI( i2, i1, i, pv, parity)		\
do {							\
   const GLubyte flags = cullmask[i];			\
							\
   if (!(flags & PRIM_NOT_CULLED))			\
      continue;						\
							\
   RVB_COLOR(pv);					\
   if (flags & PRIM_ANY_CLIP) {				\
      if (parity) {					\
	 vlist[0] = i1;					\
	 vlist[1] = i2;					\
	 vlist[2] = i;					\
      } else {						\
	 vlist[0] = i2;					\
	 vlist[1] = i1;					\
	 vlist[2] = i;					\
      }							\
      fxRenderClippedTriangle(VB,3,vlist);		\
   } else if (parity)					\
     grDrawTriangle(&gWin[i1],&gWin[i2],&gWin[i]);	\
   else							\
     grDrawTriangle(&gWin[i2],&gWin[i1],&gWin[i]);	\
							\
} while (0)


#define RENDER_QUAD(i3, i2, i1, i, pv)		\
do {							\
   const GLubyte flags = cullmask[i];			\
							\
   if (!(flags & PRIM_NOT_CULLED))			\
      continue;						\
							\
   RVB_COLOR(pv);					\
   if (flags&PRIM_ANY_CLIP) {				\
      vlist[0] = i3;					\
      vlist[1] = i2;					\
      vlist[2] = i1;					\
      vlist[3] = i;					\
      fxRenderClippedTriangle(VB,4,vlist);		\
   } else {						\
      grDrawTriangle(&gWin[i3],&gWin[i2],&gWin[i]);	\
      grDrawTriangle(&gWin[i2],&gWin[i1],&gWin[i]);	\
   }							\
} while (0)





#define LOCAL_VARS					\
   fxMesaContext fxMesa=(fxMesaContext)VB->ctx->DriverCtx;	\
   const GrVertex *gWin=(GrVertex *)VB->driver_data;	\
   const GLubyte *cullmask = VB->CullMask;		\
   GLuint vlist[VB_SIZE];				\
   (void) vlist; (void) fxMesa;


#define INIT(x)  

#define TAB    render_vb_smooth_culled
#define TAG(x) x##_fx_smooth_culled
#undef  RVB_COLOR
#define RVB_COLOR(x)
#define PRESERVE_VB_DEFS
#include "render_tmp.h"

#define TAB    render_vb_flat_culled
#define TAG(x) x##_fx_flat_culled
#undef  RVB_COLOR
#define RVB_COLOR(pv)  FX_VB_COLOR(fxMesa, VB->ColorPtr->data[pv])

#include "render_tmp.h"
















/* Direct, with the possibility of clipping.
 */ 
#define RENDER_POINTS( start, count )		\
do {						\
   GLuint e;					\
   GLubyte *clipmask = VB->ClipMask;		\
   GrVertex *gWin=(GrVertex *)VB->driver_data;	\
   for(e=start;e<=count;e++)			\
     if(clipmask[elt[e]]==0)			\
        grDrawPoint(&gWin[elt[e]]);		\
} while (0)

#define RENDER_LINE( i1, i )			\
do {						\
   GLuint e1 = elt[i1], e = elt[i];		\
   fxRenderClippedLine( VB, e1, e );		\
} while (0)

#define RENDER_TRI( i2, i1, i, pv, parity)		\
do {							\
  GLuint e2 = elt[i2], e1 = elt[i1], e = elt[i];	\
  if (parity) { GLuint t=e2; e2=e1; e1=t; }		\
  fxRenderClippedTriangle2(VB,e2,e1,e);			\
} while (0)

#define RENDER_QUAD( i3, i2, i1, i, pv)				\
do {								\
  GLuint e3 = elt[i3], e2 = elt[i2], e1 = elt[i1], e = elt[i];	\
  fxRenderClippedTriangle2(VB,e3,e2,e);				\
  fxRenderClippedTriangle2(VB,e2,e1,e);				\
} while (0)

#define LOCAL_VARS						\
    const GLuint *elt = VB->Elt;			

#define INIT(x)  

#define TAB    render_vb_smooth_indirect_clipped
#define TAG(x) x##_fx_smooth_indirect_clipped
#include "render_tmp.h"







/* Direct, and no clipping required.
 */
#define RENDER_POINTS( start, count )		\
do {						\
   GLuint e;					\
   for(e=start;e<=count;e++)			\
     grDrawPoint(&gWin[elt[e]]);			\
} while (0)

#define RENDER_LINE( i1, i )			\
do {						\
   GLuint e1 = elt[i1], e = elt[i];		\
   grDrawLine( &gWin[e1], &gWin[e] );		\
} while (0)


#define RENDER_TRI( i2, i1, i, pv, parity)		\
do {							\
  GLuint e2 = elt[i2], e1 = elt[i1], e = elt[i];	\
  if (parity)						\
    grDrawTriangle(&gWin[e1],&gWin[e2],&gWin[e]);		\
  else							\
    grDrawTriangle(&gWin[e2],&gWin[e1],&gWin[e]);		\
} while (0)


#define RENDER_QUAD( i3, i2, i1, i, pv)				\
do {								\
  GLuint e3 = elt[i3], e2 = elt[i2], e1 = elt[i1], e = elt[i];	\
  grDrawTriangle(&gWin[e3],&gWin[e2],&gWin[e]);			\
  grDrawTriangle(&gWin[e2],&gWin[e1],&gWin[e]);			\
} while (0)

#define LOCAL_VARS						\
    GrVertex *gWin=(GrVertex *)VB->driver_data;		\
    const GLuint *elt = VB->Elt;
#define INIT(x)  

#define TAB    render_vb_smooth_indirect
#define TAG(x) x##_fx_smooth_indirect
#include "render_tmp.h"





/* Direct in this context means that triangles, lines, points can be
 * rendered simply by calling grDrawTriangle, etc., without any
 * additional setup (such as calling grConstantColor).  We also use a
 * 'safe' set of clipping routines which don't require write-access to
 * the arrays in the vertex buffer, and don't care about array
 * stride.  
 *
 * Thus there is no call to gl_import_arrays() in this function.  
 *
 * This safe clipping should be generalized to call driver->trianglefunc
 * under the appropriate conditions.
 *
 * We don't handle texcoord-4 in the safe clip routines - maybe we should.
 *
 */
void fxDDRenderElementsDirect( struct vertex_buffer *VB,
			       GLenum prim,
			       GLuint *elt,
			       GLuint nr )
{
  GLcontext *ctx = VB->ctx;
  struct vertex_buffer *saved_vb = ctx->VB;
  GLuint *saved_elt = VB->Elt;

  render_func func;

  if (VB->ClipOrMask)
    func = render_vb_smooth_indirect_clipped[prim];
  else 
    func = render_vb_smooth_indirect[prim];
    
  ctx->VB = VB;			/* kludge */
  VB->Elt = elt;		/* another kludge */

/*    printf("fxDDRenderElementsDirect: %s/%d %d-%d\n",  */
/*  	 gl_prim_name[prim], 0, 0, nr); */

  func( VB, 0, nr, 0 );

  ctx->VB = saved_vb;
  VB->Elt = saved_elt;
}




static render_func *fxDDRenderVBSmooth_tables[3] = {
  0,
  render_vb_smooth_culled,
  render_vb_smooth_raw
};

static render_func *fxDDRenderVBFlat_tables[3] = {
  0,
  render_vb_flat_culled,
  render_vb_flat_raw
};


static render_func *null_tables[3] = {
  0,
  0,
  0
};


void fxDDRenderInit( GLcontext *ctx )
{
   render_init_fx_smooth_raw();
   render_init_fx_smooth_culled();
   render_init_fx_flat_raw();
   render_init_fx_flat_culled();
   render_init_fx_smooth_indirect();
   render_init_fx_smooth_indirect_clipped();
}



render_func **fxDDChooseRenderVBTables(GLcontext *ctx)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;

  if(ctx->Polygon.CullFlag) {
    switch(ctx->Polygon.CullFaceMode) {
    case GL_BACK:
      if(ctx->Polygon.FrontFace==GL_CCW)
	grCullMode(GR_CULL_NEGATIVE);
      else
	grCullMode(GR_CULL_POSITIVE);
      break;
    case GL_FRONT:
      if(ctx->Polygon.FrontFace==GL_CCW)
	grCullMode(GR_CULL_POSITIVE);
      else
	grCullMode(GR_CULL_NEGATIVE);
      break;
    case GL_FRONT_AND_BACK:
      grCullMode(GR_CULL_DISABLE);
      return null_tables;
    default:
      ; /*nothing*/
    }
  } else
    grCullMode(GR_CULL_DISABLE);

  if (ctx->IndirectTriangles)
    return null_tables;

  switch (fxMesa->render_index) {
  case FX_FLAT:
    return fxDDRenderVBFlat_tables;
  case 0:
    return fxDDRenderVBSmooth_tables; 
  default:
    return null_tables;
  }
}




#else


/*
 * Need this to provide at least one external definition.
 */

int gl_fx_dummy_function_render(void)
{
  return 0;
}

#endif  /* FX */
