/* -*- mode: C; tab-width:8;  -*-

             fxsetup.c - 3Dfx VooDoo rendering mode setup functions
*/

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * See the file fxapi.c for more informations about authors
 *
 */

#if defined(FX)

#include "fxdrv.h"

static void fxTexValidate(GLcontext *ctx, struct gl_texture_object *tObj)
{
  tfxTexInfo *ti=(tfxTexInfo *)tObj->DriverData;
  GLint minl,maxl;

  if (MESA_VERBOSE&VERBOSE_DRIVER) {
     fprintf(stderr,"fxmesa: fxTexValidate(...) Start\n");
  }

  if(ti->validated) {
     if (MESA_VERBOSE&VERBOSE_DRIVER) {
	fprintf(stderr,"fxmesa: fxTexValidate(...) End (validated=GL_TRUE)\n");
     }
    return;
  }

  minl=ti->minLevel=tObj->BaseLevel;
  maxl=ti->maxLevel=MIN2(tObj->MaxLevel,tObj->Image[0]->MaxLog2);

  fxTexGetInfo(tObj->Image[minl]->Width,tObj->Image[minl]->Height,
	       &(ti->info.largeLod),&(ti->info.aspectRatio),
	       &(ti->sScale),&(ti->tScale),
	       NULL,NULL);


/*   printf("sScale: %f tscale: %f\n", ti->sScale, ti->tScale); */

  if((tObj->MinFilter!=GL_NEAREST) && (tObj->MinFilter!=GL_LINEAR))
    fxTexGetInfo(tObj->Image[maxl]->Width,tObj->Image[maxl]->Height,
		 &(ti->info.smallLod),NULL,
		 NULL,NULL,
		 NULL,NULL);
  else
    ti->info.smallLod=ti->info.largeLod;

  fxTexGetFormat(tObj->Image[minl]->Format,&(ti->info.format),&(ti->baseLevelInternalFormat));

  ti->validated=GL_TRUE;

  ti->info.data=NULL;

  if (MESA_VERBOSE&VERBOSE_DRIVER) {
     fprintf(stderr,"fxmesa: fxTexValidate(...) End\n");
  }
}

GLuint fxGetTexSetConfiguration(GLcontext *ctx,
				struct gl_texture_object *tObj0,
				struct gl_texture_object *tObj1)
{
  GLuint unitsmode=0;
  GLuint envmode=0;
  GLuint ifmt=0;

  if((ctx->Light.ShadeModel==GL_SMOOTH) ||
     (ctx->Point.SmoothFlag) ||
     (ctx->Line.SmoothFlag) ||
     (ctx->Polygon.SmoothFlag))
    unitsmode|=FX_UM_ALPHA_ITERATED;
  else
    unitsmode|=FX_UM_ALPHA_CONSTANT;

  if(ctx->Light.ShadeModel==GL_SMOOTH)
    unitsmode|=FX_UM_COLOR_ITERATED;
  else
    unitsmode|=FX_UM_COLOR_CONSTANT;

  if(tObj0) {
    tfxTexInfo *ti0=(tfxTexInfo *)tObj0->DriverData;

    switch(ti0->baseLevelInternalFormat) {
    case GL_ALPHA:
      ifmt|=FX_UM_E0_ALPHA;
      break;
    case GL_LUMINANCE:
      ifmt|=FX_UM_E0_LUMINANCE;
      break;
    case GL_LUMINANCE_ALPHA:
      ifmt|=FX_UM_E0_LUMINANCE_ALPHA;
      break;
    case GL_INTENSITY:
      ifmt|=FX_UM_E0_INTENSITY;
      break;
    case GL_RGB:
      ifmt|=FX_UM_E0_RGB;
      break;
    case GL_RGBA:
      ifmt|=FX_UM_E0_RGBA;
      break;
    }

    switch(ctx->Texture.Unit[0].EnvMode) {
    case GL_DECAL:
      envmode|=FX_UM_E0_DECAL;
      break;
    case GL_MODULATE:
      envmode|=FX_UM_E0_MODULATE;
      break;
    case GL_REPLACE:
      envmode|=FX_UM_E0_REPLACE;
      break;
    case GL_BLEND:
      envmode|=FX_UM_E0_BLEND;
      break;
    default:
      /* do nothing */
      break;
    }
  }

  if(tObj1) {
    tfxTexInfo *ti1=(tfxTexInfo *)tObj1->DriverData;

    switch(ti1->baseLevelInternalFormat) {
    case GL_ALPHA:
      ifmt|=FX_UM_E1_ALPHA;
      break;
    case GL_LUMINANCE:
      ifmt|=FX_UM_E1_LUMINANCE;
      break;
    case GL_LUMINANCE_ALPHA:
      ifmt|=FX_UM_E1_LUMINANCE_ALPHA;
      break;
    case GL_INTENSITY:
      ifmt|=FX_UM_E1_INTENSITY;
      break;
    case GL_RGB:
      ifmt|=FX_UM_E1_RGB;
      break;
    case GL_RGBA:
      ifmt|=FX_UM_E1_RGBA;
      break;
    default:
      /* do nothing */
      break;
    }

    switch(ctx->Texture.Unit[1].EnvMode) {
    case GL_DECAL:
      envmode|=FX_UM_E1_DECAL;
      break;
    case GL_MODULATE:
      envmode|=FX_UM_E1_MODULATE;
      break;
    case GL_REPLACE:
      envmode|=FX_UM_E1_REPLACE;
      break;
    case GL_BLEND:
      envmode|=FX_UM_E1_BLEND;
      break;
    default:
      /* do nothing */
      break;
    }
  }

  unitsmode|=(ifmt | envmode);

  return unitsmode;
}

/************************************************************************/
/************************* Rendering Mode SetUp *************************/
/************************************************************************/

/************************* Single Texture Set ***************************/

static void fxSetupSingleTMU(fxMesaContext fxMesa, struct gl_texture_object *tObj)
{
  tfxTexInfo *ti=(tfxTexInfo *)tObj->DriverData;

  if(!ti->tmi.isInTM) {
    if(ti->LODblend)
      fxTMMoveInTM(fxMesa,tObj,FX_TMU_SPLIT);
    else {
      if(fxMesa->haveTwoTMUs) {
	if(fxMesa->freeTexMem[FX_TMU0]>grTexTextureMemRequired(GR_MIPMAPLEVELMASK_BOTH,&(ti->info)))
	  fxTMMoveInTM(fxMesa,tObj,FX_TMU0);
	else
	  fxTMMoveInTM(fxMesa,tObj,FX_TMU1);
      } else
	fxTMMoveInTM(fxMesa,tObj,FX_TMU0);
    }
  }

  if(ti->LODblend && ti->tmi.whichTMU == FX_TMU_SPLIT) {
    if((ti->info.format==GR_TEXFMT_P_8) && (!fxMesa->haveGlobalPaletteTexture)) {
       if (MESA_VERBOSE&VERBOSE_DRIVER) {
	  fprintf(stderr,"fxmesa: uploading texture palette\n");
       }
      grTexDownloadTable(GR_TMU0,GR_TEXTABLE_PALETTE,&(ti->palette));
      grTexDownloadTable(GR_TMU1,GR_TEXTABLE_PALETTE,&(ti->palette));
    }

    grTexClampMode(GR_TMU0,ti->sClamp,ti->tClamp);
    grTexClampMode(GR_TMU1,ti->sClamp,ti->tClamp);
    grTexFilterMode(GR_TMU0,ti->minFilt,ti->maxFilt);
    grTexFilterMode(GR_TMU1,ti->minFilt,ti->maxFilt);
    grTexMipMapMode(GR_TMU0,ti->mmMode,ti->LODblend);
    grTexMipMapMode(GR_TMU1,ti->mmMode,ti->LODblend);

    grTexSource(GR_TMU0,ti->tmi.tm[FX_TMU0]->startAddress,
		GR_MIPMAPLEVELMASK_ODD,&(ti->info));
    grTexSource(GR_TMU1,ti->tmi.tm[FX_TMU1]->startAddress,
		GR_MIPMAPLEVELMASK_EVEN,&(ti->info));
  } else {
    if((ti->info.format==GR_TEXFMT_P_8) && (!fxMesa->haveGlobalPaletteTexture)) {
       if (MESA_VERBOSE&VERBOSE_DRIVER) {
	  fprintf(stderr,"fxmesa: uploading texture palette\n");
       }
      grTexDownloadTable(ti->tmi.whichTMU,GR_TEXTABLE_PALETTE,&(ti->palette));
    }

    /* KW: The alternative is to do the download to the other tmu.  If
     * we get to this point, I think it means we are thrashing the
     * texture memory, so perhaps it's not a good idea.  
     */
    if (ti->LODblend && (MESA_VERBOSE&VERBOSE_DRIVER))
       fprintf(stderr, "fxmesa: not blending texture - only on one tmu\n");

    grTexClampMode(ti->tmi.whichTMU,ti->sClamp,ti->tClamp);
    grTexFilterMode(ti->tmi.whichTMU,ti->minFilt,ti->maxFilt);
    grTexMipMapMode(ti->tmi.whichTMU,ti->mmMode,FXFALSE);

    grTexSource(ti->tmi.whichTMU,ti->tmi.tm[ti->tmi.whichTMU]->startAddress,
		GR_MIPMAPLEVELMASK_BOTH,&(ti->info));
  }
}

static void fxSelectSingleTMUSrc(fxMesaContext fxMesa, GLint tmu, FxBool LODblend)
{
   if (MESA_VERBOSE&VERBOSE_DRIVER) {
      fprintf(stderr,"fxmesa: fxSelectSingleTMUSrc(%d,%d)\n",tmu,LODblend);
   }

  if(LODblend) {
    grTexCombine(GR_TMU0,
		 GR_COMBINE_FUNCTION_BLEND,
		 GR_COMBINE_FACTOR_ONE_MINUS_LOD_FRACTION,
		 GR_COMBINE_FUNCTION_BLEND,
		 GR_COMBINE_FACTOR_ONE_MINUS_LOD_FRACTION,
		 FXFALSE,FXFALSE);

    grTexCombine(GR_TMU1,
		 GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		 GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		 FXFALSE,FXFALSE);

    fxMesa->tmuSrc=FX_TMU_SPLIT;
  } else {
    if(tmu==FX_TMU0) {
      grTexCombine(GR_TMU0,
		   GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		   GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		   FXFALSE,FXFALSE);
      
      fxMesa->tmuSrc=FX_TMU0;
    } else {
      grTexCombine(GR_TMU1,
		   GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		   GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		   FXFALSE,FXFALSE);
    
      /* GR_COMBINE_FUNCTION_SCALE_OTHER doesn't work ?!? */
    
      grTexCombine(GR_TMU0,
		   GR_COMBINE_FUNCTION_BLEND,GR_COMBINE_FACTOR_ONE,
		   GR_COMBINE_FUNCTION_BLEND,GR_COMBINE_FACTOR_ONE,
		   FXFALSE,FXFALSE);
    
      fxMesa->tmuSrc=FX_TMU1;
    }
  }
}

static void fxSetupTextureSingleTMU(GLcontext *ctx, GLuint textureset)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;
  GrCombineLocal_t localc,locala;
  GLuint unitsmode;
  GLint ifmt;
  tfxTexInfo *ti;
  struct gl_texture_object *tObj=ctx->Texture.Unit[textureset].Current2D;

  if (MESA_VERBOSE&VERBOSE_DRIVER) {
     fprintf(stderr,"fxmesa: fxSetupTextureSingleTMU(...) Start\n");
  }

  ti=(tfxTexInfo *)tObj->DriverData;

  fxTexValidate(ctx,tObj);

  fxSetupSingleTMU(fxMesa,tObj);

  if(fxMesa->tmuSrc!=ti->tmi.whichTMU)
    fxSelectSingleTMUSrc(fxMesa,ti->tmi.whichTMU,ti->LODblend);

  if(textureset==0)
    unitsmode=fxGetTexSetConfiguration(ctx,tObj,NULL);
  else
    unitsmode=fxGetTexSetConfiguration(ctx,NULL,tObj);

  if(fxMesa->lastUnitsMode==unitsmode)
    return;

  fxMesa->lastUnitsMode=unitsmode;

  fxMesa->stw_hint_state = 0;
  grHints(GR_HINT_STWHINT,0);

  ifmt=ti->baseLevelInternalFormat;

  if(unitsmode & FX_UM_ALPHA_ITERATED)
    locala=GR_COMBINE_LOCAL_ITERATED;
  else
    locala=GR_COMBINE_LOCAL_CONSTANT;

  if(unitsmode & FX_UM_COLOR_ITERATED)
    localc=GR_COMBINE_LOCAL_ITERATED;
  else
    localc=GR_COMBINE_LOCAL_CONSTANT;

  switch(ctx->Texture.Unit[textureset].EnvMode) {
  case GL_DECAL:
    grAlphaCombine(GR_COMBINE_FUNCTION_LOCAL,
		   GR_COMBINE_FACTOR_NONE,
		   locala,
		   GR_COMBINE_OTHER_NONE,
		   FXFALSE);

    grColorCombine(GR_COMBINE_FUNCTION_BLEND,
		   GR_COMBINE_FACTOR_TEXTURE_ALPHA,
		   localc,
		   GR_COMBINE_OTHER_TEXTURE,
		   FXFALSE);
    break;
  case GL_MODULATE:
    grAlphaCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		   GR_COMBINE_FACTOR_LOCAL,
		   locala,
		   GR_COMBINE_OTHER_TEXTURE,
		   FXFALSE);

    if(ifmt==GL_ALPHA)
      grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
		     GR_COMBINE_FACTOR_NONE,
		     localc,
		     GR_COMBINE_OTHER_NONE,
		     FXFALSE);
    else
      grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		     GR_COMBINE_FACTOR_LOCAL,
		     localc,
		     GR_COMBINE_OTHER_TEXTURE,
		     FXFALSE);
    break;
  case GL_BLEND:
#ifndef FX_SILENT
    fprintf(stderr,"fx Driver: GL_BLEND not yet supported\n");
#endif
    /* TO DO (I think that the Voodoo Graphics isn't able to support GL_BLEND) */
    break;
  case GL_REPLACE:
    if((ifmt==GL_RGB) || (ifmt==GL_LUMINANCE))
      grAlphaCombine(GR_COMBINE_FUNCTION_LOCAL,
		     GR_COMBINE_FACTOR_NONE,
		     locala,
		     GR_COMBINE_OTHER_NONE,
		     FXFALSE);
    else
      grAlphaCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		     GR_COMBINE_FACTOR_ONE,
		     locala,
		     GR_COMBINE_OTHER_TEXTURE,
		     FXFALSE);
    
    if(ifmt==GL_ALPHA)
      grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
		     GR_COMBINE_FACTOR_NONE,
		     localc,
		     GR_COMBINE_OTHER_NONE,
		     FXFALSE);
    else
      grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		     GR_COMBINE_FACTOR_ONE,
		     localc,
		     GR_COMBINE_OTHER_TEXTURE,
		     FXFALSE);
    break;
  default:
#ifndef FX_SILENT
    fprintf(stderr,"fx Driver: %x Texture.EnvMode not yet supported\n",ctx->Texture.Unit[textureset].EnvMode);
#endif
    break;
  }

  if (MESA_VERBOSE&VERBOSE_DRIVER) {
     fprintf(stderr,"fxmesa: fxSetupTextureSingleTMU(...) End\n");
  }
}

/************************* Double Texture Set ***************************/

void fxSetupDoubleTMU(fxMesaContext fxMesa, struct gl_texture_object *tObj0,
		      struct gl_texture_object *tObj1)
{
#define T0_NOT_IN_TMU  0x01
#define T1_NOT_IN_TMU  0x02
#define T0_IN_TMU0     0x04
#define T1_IN_TMU0     0x08
#define T0_IN_TMU1     0x10
#define T1_IN_TMU1     0x20

  tfxTexInfo *ti0=(tfxTexInfo *)tObj0->DriverData;
  tfxTexInfo *ti1=(tfxTexInfo *)tObj1->DriverData;
  GLuint tstate=0;

  if (MESA_VERBOSE&VERBOSE_DRIVER) {
     fprintf(stderr,"fxmesa: fxSetupDoubleTMU(...)\n");
  }

  if(ti0->tmi.isInTM) {
    if(ti0->tmi.whichTMU==FX_TMU0)
      tstate|=T0_IN_TMU0;
    else if(ti0->tmi.whichTMU==FX_TMU1)
      tstate|=T0_IN_TMU1;
    else {
      fxTMMoveOutTM(fxMesa,tObj0);
      tstate|=T0_NOT_IN_TMU;
    }
  } else
    tstate|=T0_NOT_IN_TMU;

  if(ti1->tmi.isInTM) {
    if(ti1->tmi.whichTMU==FX_TMU0)
      tstate|=T1_IN_TMU0;
    else if(ti1->tmi.whichTMU==FX_TMU1)
      tstate|=T1_IN_TMU1;
    else {
      fxTMMoveOutTM(fxMesa,tObj1);
      tstate|=T1_NOT_IN_TMU;
    }
  } else
    tstate|=T1_NOT_IN_TMU;

  ti0->tmi.lastTimeUsed=fxMesa->texBindNumber;
  ti1->tmi.lastTimeUsed=fxMesa->texBindNumber;

  /* Move texture maps in TMUs */ 

  switch(tstate) {
  case (T0_IN_TMU0 | T1_IN_TMU0):
    fxTMMoveOutTM(fxMesa,tObj1);

    fxTMMoveInTM(fxMesa,tObj1,FX_TMU1);
    break;

  case (T0_IN_TMU1 | T1_IN_TMU1):
    fxTMMoveOutTM(fxMesa,tObj0);

    fxTMMoveInTM(fxMesa,tObj0,FX_TMU0);
    break;

  case (T0_NOT_IN_TMU | T1_NOT_IN_TMU):
    fxTMMoveInTM(fxMesa,tObj0,FX_TMU0);
    fxTMMoveInTM(fxMesa,tObj1,FX_TMU1);
    break;

    /*** T0/T1 ***/

  case (T0_NOT_IN_TMU | T1_IN_TMU0):
    fxTMMoveInTM(fxMesa,tObj0,FX_TMU1);
    break;

  case (T0_NOT_IN_TMU | T1_IN_TMU1):
    fxTMMoveInTM(fxMesa,tObj0,FX_TMU0);
    break;

  case (T0_IN_TMU0 | T1_NOT_IN_TMU):
    fxTMMoveInTM(fxMesa,tObj1,FX_TMU1);
    break;

  case (T0_IN_TMU1 | T1_NOT_IN_TMU):
    fxTMMoveInTM(fxMesa,tObj1,FX_TMU0);
    break;

    /*** Best Case ***/

  case (T0_IN_TMU1 | T1_IN_TMU0):
  case (T0_IN_TMU0 | T1_IN_TMU1):
    break;

  default:
    fprintf(stderr,"fx Driver: internal error in fxSetupDoubleTMU()\n");
    fxCloseHardware();
    exit(-1);
    break;
  }

  if(!fxMesa->haveGlobalPaletteTexture) {
    if(ti0->info.format==GR_TEXFMT_P_8) {
       if (MESA_VERBOSE&VERBOSE_DRIVER) {
	  fprintf(stderr,"fxmesa: uploading texture palette TMU0\n");
       }
      grTexDownloadTable(ti0->tmi.whichTMU,GR_TEXTABLE_PALETTE,&(ti0->palette));
    }

    if(ti1->info.format==GR_TEXFMT_P_8) {
       if (MESA_VERBOSE&VERBOSE_DRIVER) {
	  fprintf(stderr,"fxmesa: uploading texture palette TMU1\n");
       }
      grTexDownloadTable(ti1->tmi.whichTMU,GR_TEXTABLE_PALETTE,&(ti1->palette));
    }
  }

  grTexClampMode(ti0->tmi.whichTMU,ti0->sClamp,ti0->tClamp);
  grTexFilterMode(ti0->tmi.whichTMU,ti0->minFilt,ti0->maxFilt);
  grTexMipMapMode(ti0->tmi.whichTMU,ti0->mmMode,FXFALSE);
  grTexSource(ti0->tmi.whichTMU,ti0->tmi.tm[ti0->tmi.whichTMU]->startAddress,
	      GR_MIPMAPLEVELMASK_BOTH,&(ti0->info));

  grTexClampMode(ti1->tmi.whichTMU,ti1->sClamp,ti1->tClamp);
  grTexFilterMode(ti1->tmi.whichTMU,ti1->minFilt,ti1->maxFilt);
  grTexMipMapMode(ti1->tmi.whichTMU,ti1->mmMode,FXFALSE);
  grTexSource(ti1->tmi.whichTMU,ti1->tmi.tm[ti1->tmi.whichTMU]->startAddress,
	      GR_MIPMAPLEVELMASK_BOTH,&(ti1->info));

#undef T0_NOT_IN_TMU
#undef T1_NOT_IN_TMU
#undef T0_IN_TMU0
#undef T1_IN_TMU0
#undef T0_IN_TMU1
#undef T1_IN_TMU1
}

static void fxSetupTextureDoubleTMU(GLcontext *ctx)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;
  GrCombineLocal_t localc,locala;
  tfxTexInfo *ti0,*ti1;
  struct gl_texture_object *tObj0=ctx->Texture.Unit[0].Current2D;
  struct gl_texture_object *tObj1=ctx->Texture.Unit[1].Current2D;
  GLuint envmode,ifmt,unitsmode;

  if (MESA_VERBOSE&VERBOSE_DRIVER) {
     fprintf(stderr,"fxmesa: fxSetupTextureDoubleTMU(...) Start\n");
  }

  ti0=(tfxTexInfo *)tObj0->DriverData;
  fxTexValidate(ctx,tObj0);

  ti1=(tfxTexInfo *)tObj1->DriverData;
  fxTexValidate(ctx,tObj1);

  fxSetupDoubleTMU(fxMesa,tObj0,tObj1);

  unitsmode=fxGetTexSetConfiguration(ctx,tObj0,tObj1);

  if(fxMesa->lastUnitsMode==unitsmode)
    return;

  fxMesa->lastUnitsMode=unitsmode;

  fxMesa->stw_hint_state |= GR_STWHINT_ST_DIFF_TMU1;
  grHints(GR_HINT_STWHINT, fxMesa->stw_hint_state);

  envmode=unitsmode & FX_UM_E_ENVMODE;
  ifmt=unitsmode & FX_UM_E_IFMT;

  if(unitsmode & FX_UM_ALPHA_ITERATED)
    locala=GR_COMBINE_LOCAL_ITERATED;
  else
    locala=GR_COMBINE_LOCAL_CONSTANT;

  if(unitsmode & FX_UM_COLOR_ITERATED)
    localc=GR_COMBINE_LOCAL_ITERATED;
  else
    localc=GR_COMBINE_LOCAL_CONSTANT;

  fxMesa->tmuSrc=FX_TMU_BOTH;
  switch(envmode) {
  case (FX_UM_E0_MODULATE | FX_UM_E1_MODULATE):
    {
      GLboolean isalpha[FX_NUM_TMU];

      if(ti0->baseLevelInternalFormat==GL_ALPHA)
	isalpha[ti0->tmi.whichTMU]=GL_TRUE;
      else
	isalpha[ti0->tmi.whichTMU]=GL_FALSE;

      if(ti1->baseLevelInternalFormat==GL_ALPHA)
	isalpha[ti1->tmi.whichTMU]=GL_TRUE;
      else
	isalpha[ti1->tmi.whichTMU]=GL_FALSE;
	
      if(isalpha[FX_TMU1])
	grTexCombine(GR_TMU1,
		     GR_COMBINE_FUNCTION_ZERO,GR_COMBINE_FACTOR_NONE,
		     GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		     FXTRUE,FXFALSE);
      else
	grTexCombine(GR_TMU1,
		     GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		     GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		     FXFALSE,FXFALSE);
    
      if(isalpha[FX_TMU0])
	grTexCombine(GR_TMU0,
		     GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_ONE,
		     GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_LOCAL,
		     FXFALSE,FXFALSE);
      else
	grTexCombine(GR_TMU0,
		     GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_LOCAL,
		     GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_LOCAL,
		     FXFALSE,FXFALSE);

      grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		     GR_COMBINE_FACTOR_LOCAL,
		     localc,
		     GR_COMBINE_OTHER_TEXTURE,
		     FXFALSE);

      grAlphaCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		     GR_COMBINE_FACTOR_LOCAL,
		     locala,
		     GR_COMBINE_OTHER_TEXTURE,
		     FXFALSE);
      break;
    }
  case (FX_UM_E0_REPLACE | FX_UM_E1_BLEND): /* Only for GLQuake */
    if(ti1->tmi.whichTMU==FX_TMU1) {
      grTexCombine(GR_TMU1,
		   GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		   GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		   FXTRUE,FXFALSE);
		  
      grTexCombine(GR_TMU0,
		   GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_LOCAL,
		   GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_LOCAL,
		   FXFALSE,FXFALSE);
    } else {
      grTexCombine(GR_TMU1,
		   GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		   GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		   FXFALSE,FXFALSE);
		  
      grTexCombine(GR_TMU0,
		   GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
		   GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
		   FXFALSE,FXFALSE);
    }
	  
    grAlphaCombine(GR_COMBINE_FUNCTION_LOCAL,
		   GR_COMBINE_FACTOR_NONE,
		   locala,
		   GR_COMBINE_OTHER_NONE,
		   FXFALSE);

    grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		   GR_COMBINE_FACTOR_ONE,
		   localc,
		   GR_COMBINE_OTHER_TEXTURE,
		   FXFALSE);
    break;
  case (FX_UM_E0_REPLACE | FX_UM_E1_MODULATE): /* Only for QuakeII */
    if(ti1->tmi.whichTMU==FX_TMU1) {
      grTexCombine(GR_TMU1,
		   GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		   GR_COMBINE_FUNCTION_ZERO,GR_COMBINE_FACTOR_NONE,
		   FXFALSE,FXTRUE);
		  
      grTexCombine(GR_TMU0,
		   GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_LOCAL,
		   GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_LOCAL,
		   FXFALSE,FXFALSE);

    } else {
      grTexCombine(GR_TMU1,
		   GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		   GR_COMBINE_FUNCTION_LOCAL,GR_COMBINE_FACTOR_NONE,
		   FXFALSE,FXFALSE);
		  
      grTexCombine(GR_TMU0,
		   GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_LOCAL,
		   GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_ONE,
		   FXFALSE,FXFALSE);
    }
	  
    if(ti0->baseLevelInternalFormat==GL_RGB)
      grAlphaCombine(GR_COMBINE_FUNCTION_LOCAL,
		     GR_COMBINE_FACTOR_NONE,
		     locala,
		     GR_COMBINE_OTHER_NONE,
		     FXFALSE);
    else
      grAlphaCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		     GR_COMBINE_FACTOR_ONE,
		     locala,
		     GR_COMBINE_OTHER_NONE,
		     FXFALSE);


    grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		   GR_COMBINE_FACTOR_ONE,
		   localc,
		   GR_COMBINE_OTHER_TEXTURE,
		   FXFALSE);
    break;
  }

  if (MESA_VERBOSE&VERBOSE_DRIVER) {
     fprintf(stderr,"fxmesa: fxSetupTextureDoubleTMU(...) End\n");
  }
}

/************************* No Texture ***************************/

static void fxSetupTextureNone(GLcontext *ctx)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;
  GrCombineLocal_t localc,locala;

  if (MESA_VERBOSE&VERBOSE_DRIVER) {
     fprintf(stderr,"fxmesa: fxSetupTextureNone(...)\n");
  }

  if((ctx->Light.ShadeModel==GL_SMOOTH) ||
     (ctx->Point.SmoothFlag) ||
     (ctx->Line.SmoothFlag) ||
     (ctx->Polygon.SmoothFlag))
    locala=GR_COMBINE_LOCAL_ITERATED;
  else
    locala=GR_COMBINE_LOCAL_CONSTANT;
  
  if(ctx->Light.ShadeModel==GL_SMOOTH)
    localc=GR_COMBINE_LOCAL_ITERATED;
  else
    localc=GR_COMBINE_LOCAL_CONSTANT;
  
  grAlphaCombine(GR_COMBINE_FUNCTION_LOCAL,
		 GR_COMBINE_FACTOR_NONE,
		 locala,
		 GR_COMBINE_OTHER_NONE,
		 FXFALSE);

  grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
		 GR_COMBINE_FACTOR_NONE,
		 localc,
		 GR_COMBINE_OTHER_NONE,
		 FXFALSE);

  fxMesa->lastUnitsMode=FX_UM_NONE;
}

/************************************************************************/
/************************** Texture Mode SetUp **************************/
/************************************************************************/

void fxSetupTexture(GLcontext *ctx)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;
  GLuint tex2Denabled;

  if (MESA_VERBOSE&VERBOSE_DRIVER) {
     fprintf(stderr,"fxmesa: fxSetupTexture(...)\n");
  }

  /* Texture Combine, Color Combine and Alpha Combine */

  tex2Denabled=0;
  if((ctx->Texture.Enabled & TEXTURE0_2D) &&
     !(ctx->Texture.Enabled & TEXTURE0_3D) &&
     (ctx->Texture.Unit[0].Current2D->Complete))
    tex2Denabled|=TEXTURE0_2D;

  if(fxMesa->haveTwoTMUs &&
     (ctx->Texture.Enabled & TEXTURE1_2D) &&
     !(ctx->Texture.Enabled & TEXTURE1_3D) &&
     (ctx->Texture.Unit[1].Current2D->Complete))
    tex2Denabled|=TEXTURE1_2D;
  
  switch(tex2Denabled) {
  case TEXTURE0_2D:
    fxSetupTextureSingleTMU(ctx,0);
    break;
  case TEXTURE1_2D:
    fxSetupTextureSingleTMU(ctx,1);
    break;
  case (TEXTURE0_2D|TEXTURE1_2D):
    fxSetupTextureDoubleTMU(ctx);
    break;
  default:
    fxSetupTextureNone(ctx);
    break;
  }
}

/************************************************************************/
/**************************** Blend SetUp *******************************/
/************************************************************************/

/* XXX consider supporting GL_INGR_blend_func_separate */
void fxDDBlendFunc(GLcontext *ctx, GLenum sfactor, GLenum dfactor)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;
  tfxUnitsState *us=&fxMesa->unitsState;
  GrAlphaBlendFnc_t sfact,dfact,asfact,adfact;

  /* From the Glide documentation:
     For alpha source and destination blend function factor
     parameters, Voodoo Graphics supports only
     GR_BLEND_ZERO and GR_BLEND_ONE.
  */

  switch(sfactor) {
  case GL_ZERO:
    asfact=sfact=GR_BLEND_ZERO;
    break;
  case GL_ONE:
    asfact=sfact=GR_BLEND_ONE;
    break;
  case GL_DST_COLOR:
    sfact=GR_BLEND_DST_COLOR;
    asfact=GR_BLEND_ONE;
    break;
  case GL_ONE_MINUS_DST_COLOR:
    sfact=GR_BLEND_ONE_MINUS_DST_COLOR;
    asfact=GR_BLEND_ONE;
    break;
  case GL_SRC_ALPHA:
    sfact=GR_BLEND_SRC_ALPHA;
    asfact=GR_BLEND_ONE;
    break;
  case GL_ONE_MINUS_SRC_ALPHA:
    sfact=GR_BLEND_ONE_MINUS_SRC_ALPHA;
    asfact=GR_BLEND_ONE;
    break;
  case GL_DST_ALPHA:
    sfact=GR_BLEND_DST_ALPHA;
    asfact=GR_BLEND_ONE;
    break;
  case GL_ONE_MINUS_DST_ALPHA:
    sfact=GR_BLEND_ONE_MINUS_DST_ALPHA;
    asfact=GR_BLEND_ONE;
    break;
  case GL_SRC_ALPHA_SATURATE:
    sfact=GR_BLEND_ALPHA_SATURATE;
    asfact=GR_BLEND_ONE;
    break;
  case GL_SRC_COLOR:
  case GL_ONE_MINUS_SRC_COLOR:
    /* USELESS */
    asfact=sfact=GR_BLEND_ONE;
    break;
  default:
    asfact=sfact=GR_BLEND_ONE;
    break;
  }

  if((sfact!=us->blendSrcFuncRGB) ||
     (asfact!=us->blendSrcFuncAlpha)) {
    us->blendSrcFuncRGB=sfact;
    us->blendSrcFuncAlpha=asfact;
    us->blendChanged=GL_TRUE;
  }

  switch(dfactor) {
  case GL_ZERO:
    adfact=dfact=GR_BLEND_ZERO;
    break;
  case GL_ONE:
    adfact=dfact=GR_BLEND_ONE;
    break;
  case GL_SRC_COLOR:
    dfact=GR_BLEND_SRC_COLOR;
    adfact=GR_BLEND_ZERO;
    break;
  case GL_ONE_MINUS_SRC_COLOR:
    dfact=GR_BLEND_ONE_MINUS_SRC_COLOR;
    adfact=GR_BLEND_ZERO;
    break;
  case GL_SRC_ALPHA:
    dfact=GR_BLEND_SRC_ALPHA;
    adfact=GR_BLEND_ZERO;
    break;
  case GL_ONE_MINUS_SRC_ALPHA:
    dfact=GR_BLEND_ONE_MINUS_SRC_ALPHA;
    adfact=GR_BLEND_ZERO;
    break;
  case GL_DST_ALPHA:
    dfact=GR_BLEND_DST_ALPHA;
    adfact=GR_BLEND_ZERO;
    break;
  case GL_ONE_MINUS_DST_ALPHA:
    dfact=GR_BLEND_ONE_MINUS_DST_ALPHA;
    adfact=GR_BLEND_ZERO;
    break;
  case GL_SRC_ALPHA_SATURATE:
  case GL_DST_COLOR:
  case GL_ONE_MINUS_DST_COLOR:
    /* USELESS */
    adfact=dfact=GR_BLEND_ZERO;
    break;
  default:
    adfact=dfact=GR_BLEND_ZERO;
    break;
  }

  if((dfact!=us->blendDstFuncRGB) ||
     (adfact!=us->blendDstFuncAlpha)) {
    us->blendDstFuncRGB=dfact;
    us->blendDstFuncAlpha=adfact;
    us->blendChanged=GL_TRUE;
  }
}

static void fxSetupBlend(GLcontext *ctx)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;
  tfxUnitsState *us=&fxMesa->unitsState;

  if(us->blendChanged) {
    if(us->blendEnabled)
      grAlphaBlendFunction(us->blendSrcFuncRGB,us->blendDstFuncRGB,
			   us->blendSrcFuncAlpha,us->blendDstFuncAlpha);
    else
      grAlphaBlendFunction(GR_BLEND_ONE,GR_BLEND_ZERO,GR_BLEND_ONE,GR_BLEND_ZERO);

    us->blendChanged=GL_FALSE;
  }
}
  
/************************************************************************/
/************************** Alpha Test SetUp ****************************/
/************************************************************************/

void fxDDAlphaFunc(GLcontext *ctx, GLenum func, GLclampf ref)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;
  tfxUnitsState *us=&fxMesa->unitsState;
  GrCmpFnc_t newfunc;

  switch(func) {
  case GL_NEVER:
    newfunc=GR_CMP_NEVER;
    break;
  case GL_LESS:
    newfunc=GR_CMP_LESS;
    break;
  case GL_EQUAL:
    newfunc=GR_CMP_EQUAL;
    break;
  case GL_LEQUAL:
    newfunc=GR_CMP_LEQUAL;
    break;
  case GL_GREATER:
    newfunc=GR_CMP_GREATER;
    break;
  case GL_NOTEQUAL:
    newfunc=GR_CMP_NOTEQUAL;
    break;
  case GL_GEQUAL:
    newfunc=GR_CMP_GEQUAL;
    break;
  case GL_ALWAYS:
    newfunc=GR_CMP_ALWAYS;
    break;
  default:
    fprintf(stderr,"fx Driver: internal error in fxDDAlphaFunc()\n");
    fxCloseHardware();
    exit(-1);
    break;
  }

  if(newfunc!=us->alphaTestFunc) {
    us->alphaTestFunc=newfunc;
    us->alphaTestChanged=GL_TRUE;
  }

  if(ctx->Color.AlphaRef!=us->alphaTestRefValue) {
    us->alphaTestRefValue=ctx->Color.AlphaRef;
    us->alphaTestChanged=GL_TRUE;
  }
}

static void fxSetupAlphaTest(GLcontext *ctx)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;
  tfxUnitsState *us=&fxMesa->unitsState;

  if(us->alphaTestChanged) {
    if(us->alphaTestEnabled) {
      grAlphaTestFunction(us->alphaTestFunc);
      grAlphaTestReferenceValue(us->alphaTestRefValue);
    } else
      grAlphaTestFunction(GR_CMP_ALWAYS);

    us->alphaTestChanged=GL_FALSE;
  }
}

/************************************************************************/
/************************** Depth Test SetUp ****************************/
/************************************************************************/

void fxDDDepthFunc(GLcontext *ctx, GLenum func)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;
  tfxUnitsState *us=&fxMesa->unitsState;
  GrCmpFnc_t dfunc;

  switch(func) {
  case GL_NEVER:
    dfunc=GR_CMP_NEVER;
    break;
  case GL_LESS:
    dfunc=GR_CMP_LESS;
    break;
  case GL_GEQUAL:
    dfunc=GR_CMP_GEQUAL;
    break;
  case GL_LEQUAL:
    dfunc=GR_CMP_LEQUAL;
    break;
  case GL_GREATER:
    dfunc=GR_CMP_GREATER;
    break;
  case GL_NOTEQUAL:
    dfunc=GR_CMP_NOTEQUAL;
    break;
  case GL_EQUAL:
    dfunc=GR_CMP_EQUAL;
    break;
  case GL_ALWAYS:
    dfunc=GR_CMP_ALWAYS;
    break;
  default:
    fprintf(stderr,"fx Driver: internal error in fxDDDepthFunc()\n");
    fxCloseHardware();
    exit(-1);
    break;
  }

  if(dfunc!=us->depthTestFunc) {
    us->depthTestFunc=dfunc;
    us->depthChanged=GL_TRUE;
  }

}

void fxDDDepthMask(GLcontext *ctx, GLboolean flag)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;
  tfxUnitsState *us=&fxMesa->unitsState;

  if(flag!=us->depthMask) {
    us->depthMask=flag;
    us->depthChanged=GL_TRUE;
  }
}

static void fxSetupDepthTest(GLcontext *ctx)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;
  tfxUnitsState *us=&fxMesa->unitsState;

  if(us->depthChanged) {
    if(us->depthTestEnabled)
      grDepthBufferFunction(us->depthTestFunc);
    else
      grDepthBufferFunction(GR_CMP_ALWAYS);

    grDepthMask(us->depthMask);
  }
}

/************************************************************************/
/**************************** Fog Mode SetUp ****************************/
/************************************************************************/

void fxFogTableGenerate(GLcontext *ctx)
{
  int i;
  float f,eyez;
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;

  for(i=0;i<GR_FOG_TABLE_SIZE;i++) {
    eyez=guFogTableIndexToW(i)*fxMesa->wScale;

    switch(ctx->Fog.Mode) {
    case GL_LINEAR:
      f=(ctx->Fog.End-eyez)/(ctx->Fog.End-ctx->Fog.Start);
      break;
    case GL_EXP:
      f=exp(-ctx->Fog.Density*eyez);  
      break;
    case GL_EXP2:
      f=exp(-ctx->Fog.Density*ctx->Fog.Density*eyez*eyez);
      break;
    default: /* That should never happen */
      f=0.0f;
      break; 
    }

    fxMesa->fogTable[i]=(GrFog_t)((1.0f-CLAMP(f,0.0f,1.0f))*255.0f);
  }
}

void fxSetupFog(GLcontext *ctx, GLboolean forceTableRebuild)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;

  if(ctx->Fog.Enabled && ctx->FogMode==FOG_FRAGMENT) {
    GLubyte col[4];
    grFogMode(GR_FOG_WITH_TABLE);

    col[0]=(unsigned int)(255*ctx->Fog.Color[0]);
    col[1]=(unsigned int)(255*ctx->Fog.Color[1]);
    col[2]=(unsigned int)(255*ctx->Fog.Color[2]); 
    col[3]=(unsigned int)(255*ctx->Fog.Color[3]);

    grFogColorValue(FXCOLOR4(col));

    if(forceTableRebuild ||
       (fxMesa->fogTableMode!=ctx->Fog.Mode) ||
       (fxMesa->fogDensity!=ctx->Fog.Density)) {
      fxFogTableGenerate(ctx);
         
      fxMesa->fogTableMode=ctx->Fog.Mode;
      fxMesa->fogDensity=ctx->Fog.Density;
    }
      
    grFogTable(fxMesa->fogTable);
  } else
    grFogMode(GR_FOG_DISABLE);
}

/************************************************************************/
/************************** Scissor Test SetUp **************************/
/************************************************************************/

static void fxSetupScissor(GLcontext *ctx)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;

  if (ctx->Scissor.Enabled) {
    int ymin, ymax;

    ymin=ctx->Scissor.Y;
    ymax=ctx->Scissor.Y+ctx->Scissor.Height;

    if (ymin<0) ymin=0;

    if (ymax>fxMesa->height) ymax=fxMesa->height;

    grClipWindow(ctx->Scissor.X, 
 		 ymin,
 		 ctx->Scissor.X+ctx->Scissor.Width, 
 		 ymax);
  } else
    grClipWindow(0,0,fxMesa->width,fxMesa->height);
}

/************************************************************************/
/****************************** DD Enable ******************************/
/************************************************************************/

void fxDDEnable(GLcontext *ctx, GLenum cap, GLboolean state)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;
  tfxUnitsState *us=&fxMesa->unitsState;

  if (MESA_VERBOSE&VERBOSE_DRIVER) {
     fprintf(stderr,"fxmesa: fxDDEnable(...)\n");
  }

  switch(cap) {
  case GL_ALPHA_TEST:
    if(state!=us->alphaTestEnabled) {
      us->alphaTestEnabled=state;
      us->alphaTestChanged=GL_TRUE;
    }
    break;
  case GL_BLEND:
    if(state!=us->blendEnabled) {
      us->blendEnabled=state;
      us->blendChanged=GL_TRUE;
    }
    break;
  case GL_DEPTH_TEST:
    if(state!=us->depthTestEnabled) {
      us->depthTestEnabled=state;
      us->depthChanged=GL_TRUE;
    }
    break;
  default:
    ;  /* XXX no-op??? */
  }    
}

/************************************************************************/
/****************************** Units SetUp *****************************/
/************************************************************************/

void fxSetupFXUnits(GLcontext *ctx)
{
  fxMesaContext fxMesa=(fxMesaContext)ctx->DriverCtx;

  if (MESA_VERBOSE&VERBOSE_DRIVER) {
     fprintf(stderr,"fxmesa: fxSetupFXUnits(...)\n");
  }

  fxSetupTexture(ctx);
  fxSetupBlend(ctx);
  fxSetupAlphaTest(ctx);
  fxSetupDepthTest(ctx);
  fxSetupFog(ctx,GL_FALSE);
  fxSetupScissor(ctx);

  grColorMask(ctx->Color.ColorMask[RCOMP] ||
	      ctx->Color.ColorMask[GCOMP] ||
	      ctx->Color.ColorMask[BCOMP],
	      ctx->Color.ColorMask[ACOMP] && fxMesa->haveAlphaBuffer);
}


#else


/*
 * Need this to provide at least one external definition.
 */

int gl_fx_dummy_function_setup(void)
{
  return 0;
}

#endif  /* FX */
