/* -*- mode: C; tab-width:8;  -*-

             fxvsetup.c - 3Dfx VooDoo vertices setup functions 
*/

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * See the file fxapi.c for more informations about authors
 *
 */

#if defined(FX)

#include "fxdrv.h"
#include "../mmath.h"
#include "fxvsetup.h"

void fxPrintSetupFlags( const char *msg, GLuint flags )
{
   fprintf(stderr, "%s: %d %s%s%s%s%s%s\n",
	  msg,
	  flags,
	  (flags & SETUP_XY) ? " xy," : "", 
	  (flags & SETUP_Z)  ? " z," : "",
	  (flags & SETUP_W)  ? " w," : "",
	  (flags & SETUP_RGBA) ? " rgba," : "",
	  (flags & SETUP_TMU0)  ? " tmu0," : "",
	  (flags & SETUP_TMU1)  ? " tmu1," : "");
}

static void project_texcoords( GrTmuVertex *tmu,
			       GLfloat (*data)[4],
			       GLubyte clipped, const GLubyte *clipmask,
			       GLuint start, GLuint count )
{
   GLuint i;

   if (!clipped) 
      for (i = start ; i < count ; i++) {
	 GLfloat ooq = 1.0 / data[i][3];
	 tmu->oow *= data[i][3];
	 tmu->sow *= ooq;
	 tmu->tow *= ooq;
	 tmu = (GrTmuVertex *)((char *)tmu + sizeof(GrVertex));
      }      
   else 
      for (i = start ; i < count ; i++) 
	 if (clipmask[i] == 0) {
	    GLfloat ooq = 1.0 / data[i][3];
	    tmu->oow *= data[i][3];
	    tmu->sow *= ooq;
	    tmu->tow *= ooq;
	    tmu = (GrTmuVertex *)((char *)tmu + sizeof(GrVertex));
	 }      
   
}

static tfxSetupFunc setupfuncs[0x40];




#define IND SETUP_XY
#define NAME fxsetupXY
#include "fxvs_tmp.h"

#define IND (SETUP_XY|SETUP_Z)
#define NAME fxsetupXYZ
#include "fxvs_tmp.h"

#define IND (SETUP_XY|SETUP_W)
#define NAME fxsetupXYW
#include "fxvs_tmp.h"

#define IND (SETUP_XY|SETUP_Z|SETUP_W)
#define NAME fxsetupXYZW
#include "fxvs_tmp.h"

#define IND (SETUP_RGBA|SETUP_XY)
#define NAME fxsetupXYRGBA
#include "fxvs_tmp.h"

#define IND (SETUP_RGBA|SETUP_XY|SETUP_Z)
#define NAME fxsetupXYZRGBA
#include "fxvs_tmp.h"

#define IND (SETUP_RGBA|SETUP_XY|SETUP_W)
#define NAME fxsetupXYWRGBA
#include "fxvs_tmp.h"

#define IND (SETUP_RGBA|SETUP_XY|SETUP_Z|SETUP_W)
#define NAME fxsetupXYZWRGBA
#include "fxvs_tmp.h"

#define IND (SETUP_TMU0|SETUP_XY|SETUP_W)
#define NAME fxsetupXYWT0
#include "fxvs_tmp.h"

#define IND (SETUP_TMU0|SETUP_XY|SETUP_Z|SETUP_W)
#define NAME fxsetupXYZWT0
#include "fxvs_tmp.h"

#define IND (SETUP_TMU1|SETUP_TMU0|SETUP_XY|SETUP_W)
#define NAME fxsetupXYWT0T1
#include "fxvs_tmp.h"

#define IND (SETUP_TMU1|SETUP_TMU0|SETUP_XY|SETUP_Z|SETUP_W)
#define NAME fxsetupXYZWT0T1
#include "fxvs_tmp.h"

#define IND (SETUP_TMU0|SETUP_RGBA|SETUP_XY|SETUP_W)
#define NAME fxsetupXYWRGBAT0
#include "fxvs_tmp.h"

#define IND (SETUP_TMU0|SETUP_RGBA|SETUP_XY|SETUP_Z|SETUP_W)
#define NAME fxsetupXYZWRGBAT0
#include "fxvs_tmp.h"

#define IND (SETUP_TMU1|SETUP_TMU0|SETUP_RGBA|SETUP_XY|SETUP_W)
#define NAME fxsetupXYWRGBAT0T1
#include "fxvs_tmp.h"

#define IND (SETUP_TMU1|SETUP_TMU0|SETUP_RGBA|SETUP_XY|SETUP_Z|SETUP_W)
#define NAME fxsetupXYZWRGBAT0T1
#include "fxvs_tmp.h"

#define IND (SETUP_RGBA)
#define NAME fxsetupRGBA
#include "fxvs_tmp.h"

#define IND (SETUP_TMU0)
#define NAME fxsetupT0
#include "fxvs_tmp.h"

#define IND (SETUP_TMU1)
#define NAME fxsetupT1
#include "fxvs_tmp.h"

#define IND (SETUP_TMU1|SETUP_TMU0)
#define NAME fxsetupT0T1
#include "fxvs_tmp.h"

#define IND (SETUP_TMU0|SETUP_RGBA)
#define NAME fxsetupRGBAT0
#include "fxvs_tmp.h"

#define IND (SETUP_TMU1|SETUP_RGBA)
#define NAME fxsetupRGBAT1
#include "fxvs_tmp.h"

#define IND (SETUP_TMU1|SETUP_TMU0|SETUP_RGBA)
#define NAME fxsetupRGBAT0T1
#include "fxvs_tmp.h"


#define IND (SETUP_W|SETUP_RGBA)
#define NAME fxsetupWRGBA
#include "fxvs_tmp.h"

#define IND (SETUP_W|SETUP_TMU0)
#define NAME fxsetupWT0
#include "fxvs_tmp.h"

#define IND (SETUP_W|SETUP_TMU1)
#define NAME fxsetupWT1
#include "fxvs_tmp.h"

#define IND (SETUP_W|SETUP_TMU1|SETUP_TMU0)
#define NAME fxsetupWT0T1
#include "fxvs_tmp.h"

#define IND (SETUP_W|SETUP_TMU0|SETUP_RGBA)
#define NAME fxsetupWRGBAT0
#include "fxvs_tmp.h"

#define IND (SETUP_W|SETUP_TMU1|SETUP_RGBA)
#define NAME fxsetupWRGBAT1
#include "fxvs_tmp.h"

#define IND (SETUP_W|SETUP_TMU1|SETUP_TMU0|SETUP_RGBA)
#define NAME fxsetupWRGBAT0T1
#include "fxvs_tmp.h"



void fxDDSetupInit()
{
   setupfuncs[SETUP_XY] = fxsetupXY;
   setupfuncs[SETUP_XY|SETUP_Z] = fxsetupXYZ;
   setupfuncs[SETUP_XY|SETUP_W] = fxsetupXYW;
   setupfuncs[SETUP_XY|SETUP_Z|SETUP_W] = fxsetupXYZW;

   setupfuncs[SETUP_RGBA|SETUP_XY] = fxsetupXYRGBA;
   setupfuncs[SETUP_RGBA|SETUP_XY|SETUP_Z] = fxsetupXYZRGBA;
   setupfuncs[SETUP_RGBA|SETUP_XY|SETUP_W] = fxsetupXYWRGBA;
   setupfuncs[SETUP_RGBA|SETUP_XY|SETUP_Z|SETUP_W] = fxsetupXYZWRGBA;

   /* If we have texture and xy then we must have w.
    * If we have texture1 and w then we must have texture 0.
    */
   setupfuncs[SETUP_TMU0|SETUP_XY|SETUP_W] = fxsetupXYWT0;
   setupfuncs[SETUP_TMU0|SETUP_XY|SETUP_Z|SETUP_W] = fxsetupXYZWT0;

   setupfuncs[SETUP_TMU1|SETUP_TMU0|SETUP_XY|SETUP_W] = fxsetupXYWT0T1;
   setupfuncs[SETUP_TMU1|SETUP_TMU0|SETUP_XY|SETUP_Z|SETUP_W] = fxsetupXYZWT0T1;

   setupfuncs[SETUP_TMU0|SETUP_RGBA|SETUP_XY|SETUP_W] = fxsetupXYWRGBAT0;
   setupfuncs[SETUP_TMU0|SETUP_RGBA|SETUP_XY|SETUP_Z|SETUP_W] = fxsetupXYZWRGBAT0;

   setupfuncs[SETUP_TMU1|SETUP_TMU0|SETUP_RGBA|SETUP_XY|SETUP_W] = fxsetupXYWRGBAT0T1;
   setupfuncs[SETUP_TMU1|SETUP_TMU0|SETUP_RGBA|SETUP_XY|SETUP_Z|SETUP_W] = fxsetupXYZWRGBAT0T1;

   /* If we don't have xy then we can't have z... w is still a possibility.
    */
   setupfuncs[SETUP_RGBA] = fxsetupRGBA;
   setupfuncs[SETUP_TMU0] = fxsetupT0;
   setupfuncs[SETUP_TMU1] = fxsetupT1;
   setupfuncs[SETUP_TMU1|SETUP_TMU0] = fxsetupT0T1;
   setupfuncs[SETUP_TMU0|SETUP_RGBA] = fxsetupRGBAT0;
   setupfuncs[SETUP_TMU1|SETUP_RGBA] = fxsetupRGBAT1;
   setupfuncs[SETUP_TMU1|SETUP_TMU0|SETUP_RGBA] = fxsetupRGBAT0T1;

   setupfuncs[SETUP_W|SETUP_RGBA] = fxsetupWRGBA;
   setupfuncs[SETUP_W|SETUP_TMU0] = fxsetupWT0;
   setupfuncs[SETUP_W|SETUP_TMU1] = fxsetupWT1;
   setupfuncs[SETUP_W|SETUP_TMU1|SETUP_TMU0] = fxsetupWT0T1;
   setupfuncs[SETUP_W|SETUP_TMU0|SETUP_RGBA] = fxsetupWRGBAT0;
   setupfuncs[SETUP_W|SETUP_TMU1|SETUP_RGBA] = fxsetupWRGBAT1;
   setupfuncs[SETUP_W|SETUP_TMU1|SETUP_TMU0|SETUP_RGBA] = fxsetupWRGBAT0T1;

}



tfxSetupFunc fxDDChooseSetupFunction(GLcontext *ctx)
{
   GLuint setupindex = SETUP_XY;
   fxMesaContext fxMesa = (fxMesaContext)ctx->DriverCtx;

   fxMesa->tmu_source[0] = 0;
   fxMesa->tmu_source[1] = 1;
   
   if (ctx->Light.ShadeModel == GL_SMOOTH && !ctx->Light.Model.TwoSide)
      setupindex |= SETUP_RGBA;

   if ((ctx->Texture.ReallyEnabled & (TEXTURE0_2D|TEXTURE0_3D)) == TEXTURE0_2D)
      setupindex |= SETUP_TMU0|SETUP_W;

   if ((ctx->Texture.ReallyEnabled & (TEXTURE1_2D|TEXTURE1_3D)) == TEXTURE1_2D)
   {
      if (setupindex & SETUP_TMU0) {
	 struct gl_texture_object *tObj=ctx->Texture.Unit[0].Current2D;
	 tfxTexInfo *ti=(tfxTexInfo *)tObj->DriverData;

	 setupindex |= SETUP_TMU1|SETUP_W;

	 if(ti->tmi.whichTMU!=FX_TMU0) {
	    fxMesa->tmu_source[0] = 1;
	    fxMesa->tmu_source[1] = 0;	    
	 }
      } else {
	 setupindex |= SETUP_TMU0|SETUP_W;
	 fxMesa->tmu_source[0] = 1;
	 fxMesa->tmu_source[1] = 0;
      }
   }

   if (ctx->Fog.Enabled && ctx->FogMode==FOG_FRAGMENT)
      setupindex |= SETUP_RGBA|SETUP_W;

   if (ctx->Depth.Test)
      setupindex |= SETUP_Z;

   if (MESA_VERBOSE)
      fxPrintSetupFlags("fxmesa: vertex setup function", setupindex); 

   fxMesa->setupindex = setupindex;
   
   return setupfuncs[setupindex];
}


static GLuint vert_flags_to_setup_flags( fxMesaContext fxMesa,
					 GLuint available )
{
   GLuint in = available; /* required & ~done; */
   GLuint ind = 0;

   /* Need to recalc texcoords if w changes - disabled SETUP_TMU units
    * will be removed in calling function.  
    *
    * Also need to redo everything if the previous VERT_WIN coordinates
    * included any clipped vertices.  I don't track this, so we just
    * redo everything available...
    */
   if (in & VERT_WIN)  {
      ind |= fxMesa->setupindex & (SETUP_XY|SETUP_W|SETUP_Z);
/*       in = required; */
   }

   if (in & VERT_TEX0_ANY)
      if (fxMesa->tmu_source[0] == 0) 
	 ind |= SETUP_W|SETUP_TMU0; 
      else
	 ind |= SETUP_W|SETUP_TMU1;

   if (in & VERT_TEX1_ANY)
      if (fxMesa->tmu_source[0] == 1) 
	 ind |= SETUP_W|SETUP_TMU0; 
      else
	 ind |= SETUP_W|SETUP_TMU1;

   if (in & VERT_RGBA)
      ind |= SETUP_W|SETUP_RGBA;

   /* remove excess elements
    */
   return ind & fxMesa->setupindex; 
}

/* Updated - when?
 *
 * Need to check that merge&render will work before allowing this to
 * happen here.  Therefore - need to know that this will be fired when
 * we get a forbidden input in the elt pipeline - and therefore need to check
 * whether we have one *now*.  Similarly need to know if state changes cause
 * size4 texcoords to be introduced.  
 */
void fxDDCheckPartialRasterSetup( GLcontext *ctx, struct gl_pipeline_stage *d )
{
   struct gl_cva *cva = &ctx->CVA;
   fxMesaContext fxMesa = (fxMesaContext)ctx->DriverCtx;
   GLuint available = ctx->RenderFlags & cva->pre.outputs;
   GLuint precalc = vert_flags_to_setup_flags( fxMesa, available );
/*    GLuint merge = fxMesa->setupindex & ~precalc; */

   /* Always recheck this stage.
    */
   d->special = PIPE_CVA_PRECALC;
   d->cva_forbidden_inputs = 0;
   fxMesa->partial_setup_index = precalc & fxMesa->setupindex;

   /* Indirect triangles still require cull information, and must be
    * rendered via the elt vb.  
    */
   if (ctx->IndirectTriangles)
      return;

   if (ctx->Texture.ReallyEnabled & 0xf) 
   {
      if (ctx->TextureMatrix[0].type == MATRIX_GENERAL ||
	  ctx->TextureMatrix[0].type == MATRIX_PERSPECTIVE)
	 return;

      if (ctx->Texture.Unit[1].TexGenEnabled & Q_BIT)
	 return;

      if (!(ctx->Array.Flags & VERT_TEX0_ANY))
	 d->cva_forbidden_inputs |= VERT_TEX0_4;
   }

   if (ctx->Texture.ReallyEnabled & 0xf0)
   {
      if (ctx->TextureMatrix[1].type == MATRIX_GENERAL ||
	  ctx->TextureMatrix[1].type == MATRIX_PERSPECTIVE)
	 return;
      
      if (ctx->Texture.Unit[1].TexGenEnabled & Q_BIT)
	 return;

      if (!(ctx->Array.Flags & VERT_TEX1_ANY))
	 d->cva_forbidden_inputs |= VERT_TEX1_4;
   }
   
   fxMesa->mergeinputs = ctx->RenderFlags & ~cva->pre.outputs;
   d->inputs = available;
   d->outputs = VERT_RAST_SETUP_PART;
   d->type = PIPE_CVA_PRECALC;
}


/* Will be different every time - no point in trying to precalc the
 * function to call.
 */
void fxDDPartialRasterSetup( struct vertex_buffer *VB, 
			     GLuint new,
			     GLuint available )
{
   fxMesaContext fxMesa = (fxMesaContext)VB->ctx->DriverCtx;
   GLuint ind = 0;

   if (new & VERT_TEX0_ANY)
      if (fxMesa->tmu_source[0] == 0) 
	 ind |= SETUP_W|SETUP_TMU0; 
      else
	 ind |= SETUP_W|SETUP_TMU1;
   
   if (new & VERT_TEX1_ANY)
      if (fxMesa->tmu_source[0] == 1) 
	 ind |= SETUP_W|SETUP_TMU0; 
      else
	 ind |= SETUP_W|SETUP_TMU1;

   if (new & VERT_RGBA)
      ind |= SETUP_W|SETUP_RGBA;

   if (new & VERT_WIN)
      ind |= SETUP_XY|SETUP_W|SETUP_Z|SETUP_TMU0|SETUP_TMU1;
   else 
      ind &= ~(fxMesa->setupdone & SETUP_W);

   ind &= fxMesa->partial_setup_index;
   ind |= fxMesa->partial_setup_index & ~fxMesa->setupdone;

/*    gl_print_vert_flags("new outputs", new ); */
/*    print_setup_flags("full partial setup", fxMesa->partial_setup_index); */
/*    print_setup_flags("part partial setup", ind ); */
/*    print_setup_flags("setup done", fxMesa->setupdone); */
/*    if (ind && !setupfuncs[ind]) */
/*       printf("no such function!\n"); */
   
   if (ind) 
      setupfuncs[ind]( VB, VB->Start, VB->Count );
   
   fxMesa->setupdone = fxMesa->partial_setup_index;
}

void fxDDInvalidateRasterSetup( struct vertex_buffer *cvaVB )
{
   fxMesaContext fxMesa = (fxMesaContext)cvaVB->ctx->DriverCtx;
   fxMesa->setupdone = 0;
}


void fxDDRegisterVB( struct vertex_buffer *VB )
{
   VB->driver_data = (GrVertex *) malloc( sizeof(GrVertex) * VB->Size );
}

void fxDDUnregisterVB( struct vertex_buffer *VB )
{
   free( VB->driver_data );
}



#else


/*
 * Need this to provide at least one external definition.
 */

int gl_fx_dummy_function_vsetup(void)
{
  return 0;
}

#endif  /* FX */
