/* $Id: render_tmp.h,v 3.6 1999/04/29 14:43:16 keithw Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * New (3.1) transformation code written by Keith Whitwell.
 */



/* Must be specially overridden for indirect case.  This should
 * get driver support for cva.
 */
void TAG(render_vb_points)( struct vertex_buffer *VB,
			    GLuint start,
			    GLuint count,
			    GLuint parity )
{
   LOCAL_VARS;
   (void) parity;
   INIT(GL_POINTS);
   RENDER_POINTS( start, count );
}

void TAG(render_vb_lines)( struct vertex_buffer *VB,
			   GLuint start,
			   GLuint count,
			   GLuint parity )
{
   GLuint j;
   GLuint *stipplecounter = &VB->ctx->StippleCounter;
   LOCAL_VARS;
   (void) parity;

   INIT(GL_LINES);
   for (j=start+1; j<count; j+=2 ) {
      RENDER_LINE( j-1, j );
      *stipplecounter = 0;
   }
}


void TAG(render_vb_line_strip)( struct vertex_buffer *VB,
				GLuint start,
				GLuint count,
				GLuint parity )
{
   GLuint j;
   LOCAL_VARS;
   (void) parity;

   INIT(GL_LINES);
   for (j=start+1; j<count; j++ ) {
      RENDER_LINE( j-1, j );
   }
   
   VB->ctx->StippleCounter = 0;	/* what is this? */
}


void TAG(render_vb_line_loop)( struct vertex_buffer *VB,
			       GLuint start,
			       GLuint count,
			       GLuint parity )
{
   GLuint i = start < VB_START ? VB_START : start + 1;	 
   LOCAL_VARS;
   (void) parity;

   INIT(GL_LINES);
   for ( ; i < count ; i++) {
      RENDER_LINE( i-1, i );
   }

   if (VB->Flag[count] & VERT_END) {
      RENDER_LINE( i-1, start );
   }   

   VB->ctx->StippleCounter = 0;
}


void TAG(render_vb_triangles)( struct vertex_buffer *VB,
			       GLuint start,
			       GLuint count,
			       GLuint parity )
{
   GLuint j;
   LOCAL_VARS;
   (void) parity;

   INIT(GL_POLYGON);
   for (j=start+2; j<count; j+=3) {
      RENDER_TRI( j-2, j-1, j, j, 0 );
   }
}



void TAG(render_vb_tri_strip)( struct vertex_buffer *VB,
			       GLuint start,
			       GLuint count,
			       GLuint parity )
{
   GLuint j;
   LOCAL_VARS;
   (void) parity;
   INIT(GL_POLYGON);
   for (j=start+2;j<count;j++,parity^=1) {
      RENDER_TRI( j-2, j-1, j, j, parity );
   }
}


void TAG(render_vb_tri_fan)( struct vertex_buffer *VB,
			     GLuint start,
			     GLuint count,
			     GLuint parity )
{
   GLuint j;
   LOCAL_VARS;
   (void) parity;
   INIT(GL_POLYGON);
   for (j=start+2;j<count;j++) {
      RENDER_TRI( start, j-1, j, j, 0 );
   }
}


void TAG(render_vb_poly)( struct vertex_buffer *VB,
			  GLuint start,
			  GLuint count,
			  GLuint parity )
{
   GLuint j;
   LOCAL_VARS;
   (void) parity;
   INIT(GL_POLYGON);
   for (j=start+2;j<count;j++) {
      RENDER_TRI( start, j-1, j, start, 0 );
   }
}


void TAG(render_vb_quads)( struct vertex_buffer *VB,
			   GLuint start,
			   GLuint count,
			   GLuint parity )
{
   GLuint j;
   LOCAL_VARS;
   (void) parity;
   INIT(GL_POLYGON);
   for (j=start+3; j<count; j+=4) {
      RENDER_QUAD( j-3, j-2, j-1, j, j );
   }
}

void TAG(render_vb_quad_strip)( struct vertex_buffer *VB,
				GLuint start,
				GLuint count,
				GLuint parity )
{
   GLuint j;
   LOCAL_VARS;
   (void) parity;
   INIT(GL_POLYGON);
   for (j=start+3;j<count;j+=2) {
      RENDER_QUAD( j-3, j-2, j, j-1, j );
   }
}


void TAG(render_init)( void )
{
   TAB[GL_POINTS]         = TAG(render_vb_points);
   TAB[GL_LINES]          = TAG(render_vb_lines);
   TAB[GL_LINE_LOOP]      = TAG(render_vb_line_loop);
   TAB[GL_LINE_STRIP]     = TAG(render_vb_line_strip);
   TAB[GL_TRIANGLES]      = TAG(render_vb_triangles);
   TAB[GL_TRIANGLE_STRIP] = TAG(render_vb_tri_strip);
   TAB[GL_TRIANGLE_FAN]   = TAG(render_vb_tri_fan);
   TAB[GL_QUADS]          = TAG(render_vb_quads);
   TAB[GL_QUAD_STRIP]     = TAG(render_vb_quad_strip);
   TAB[GL_POLYGON]        = TAG(render_vb_poly);
   TAB[GL_POLYGON+1]      = render_vb_noop;
}


#ifndef PRESERVE_VB_DEFS
#undef RENDER_TRI
#undef RENDER_QUAD
#undef RENDER_LINE
#undef RENDER_POINTS
#undef RENDER_POLYGON
#undef LOCAL_VARS
#undef INIT
#endif

#undef PRESERVE_VB_DEFS
#undef TAB
#undef TAG
